import pytest
import sqlalchemy
import transaction

from datetime import datetime, date, timedelta
from freezegun import freeze_time
from onegov.activity import ActivityCollection
from onegov.activity import ActivityFilter
from onegov.activity import Attendee
from onegov.activity import AttendeeCollection
from onegov.activity import Booking, BookingCollection
from onegov.activity import Invoice, InvoiceCollection
from onegov.activity.models.invoice_reference import FeriennetSchema
from onegov.activity.models.invoice_reference import ESRSchema
from onegov.activity import Occasion, OccasionDate
from onegov.activity import OccasionCollection
from onegov.activity import Period
from onegov.activity import PeriodCollection
from onegov.activity import PublicationRequestCollection
from onegov.activity.models import DAYS
from onegov.core.utils import Bunch
from sqlalchemy import func
from sqlalchemy.exc import IntegrityError
from psycopg2.extras import NumericRange
from pytz import utc
from uuid import uuid4


def test_add_activity(session, owner):

    collection = ActivityCollection(session)
    collection.add(
        title="Visit the Butcher",
        username=owner.username,
        lead="Come visit the butcher with us and kill your own baby pig",
        text="<h1>Babe was such an overrated movie</h1>",
        tags=('butcher', 'killing', 'blood', 'fun'),
    )

    activity = collection.by_name('visit-the-butcher')
    assert activity.title == "Visit the Butcher"
    assert activity.username == owner.username
    assert activity.reporter == owner.username
    assert activity.lead.startswith("Come visit the butcher")
    assert activity.text.startswith("<h1>Babe was such a")
    assert activity.tags == {'butcher', 'killing', 'blood', 'fun'}


def test_unique_activity(session, owner):

    collection = ActivityCollection(session)
    assert collection.get_unique_name("Möped Lads") == 'moeped-lads'

    collection.add("Möped Lads", username=owner.username)
    assert collection.get_unique_name("Möped Lads") == 'moeped-lads-1'

    collection.add("Möped Lads", username=owner.username)
    assert collection.get_unique_name("Möped Lads") == 'moeped-lads-2'


def test_activity_pagination(session, owner):

    collection = ActivityCollection(session)

    for i in range(0, 20):
        collection.add(
            title='{:02d}'.format(i),
            username=owner.username
        )

    collection = collection.page_by_index(0)
    assert [a.title for a in collection.batch] == [
        "00", "01", "02", "03", "04", "05", "06", "07", "08", "09"
    ]

    collection = collection.page_by_index(1)
    assert [a.title for a in collection.batch] == [
        "10", "11", "12", "13", "14", "15", "16", "17", "18", "19"
    ]


def test_activity_order(session, owner):

    collection = ActivityCollection(session)
    collection.add(title="Ä", username=owner.username)
    collection.add(title="B", username=owner.username)
    collection.add(title="C", username=owner.username)

    assert [a.title for a in collection.query().all()] == ["Ä", "B", "C"]


def test_activity_states(session, owner):

    c = ActivityCollection(session)
    c.add("A", username=owner.username)
    c.add("B", username=owner.username).propose()
    c.add("C", username=owner.username).propose().accept()
    c.add("D", username=owner.username).propose().accept().archive()

    c.filter.states = ('preview', )
    assert c.query().count() == 1

    c.filter.states = ('preview', 'proposed')
    assert c.query().count() == 2

    c.filter.states = ('preview', 'proposed', 'accepted')
    assert c.query().count() == 3

    c.filter.states = ('preview', 'proposed', 'accepted', 'archived')
    assert c.query().count() == 4

    c.filter.states = None
    assert c.query().count() == 4


def test_activity_used_tags(session, owner):

    c = ActivityCollection(session)

    activities = (
        c.add("Paintball", username=owner.username, tags=('sport', 'fun')),
        c.add("Dancing", username=owner.username, tags=('dance', 'fun'))
    )

    assert c.used_tags == {'sport', 'dance', 'fun'}

    c.states = ('proposed', )

    assert c.used_tags == {'sport', 'dance', 'fun'}

    activities[0].propose()

    assert c.used_tags == {'sport', 'dance', 'fun'}

    class LimitedActivityCollection(ActivityCollection):

        def query_base(self):
            return self.session.query(self.model_class).filter(
                self.model_class.state == 'proposed')

    transaction.commit()

    c = LimitedActivityCollection(session)
    assert c.used_tags == {'sport', 'fun'}


def test_occasions(session, owner):

    activities = ActivityCollection(session)
    occasions = OccasionCollection(session)
    periods = PeriodCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True
    )

    tournament = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        meeting_point="Lucerne",
        age=(6, 9),
        spots=(2, 10),
        note="Bring game-face",
        activity=activities.add("Sport", username=owner.username),
        period=period
    )

    transaction.commit()

    tournament = occasions.query().one()
    assert tournament.dates[0].start == datetime(2016, 10, 4, 11, tzinfo=utc)
    assert tournament.dates[0].end == datetime(2016, 10, 4, 12, tzinfo=utc)
    assert tournament.dates[0].timezone == "Europe/Zurich"
    assert tournament.dates[0].duration_in_seconds == 3600
    assert tournament.meeting_point == "Lucerne"
    assert tournament.note == "Bring game-face"
    assert tournament.activity_id == activities.query().first().id
    assert tournament.period_id == periods.query().first().id

    # postgres ranges are coerced into half-open intervals
    assert tournament.age.lower == 6
    assert tournament.age.upper == 10
    assert tournament.spots.lower == 2
    assert tournament.spots.upper == 11

    transaction.commit()

    tournament = occasions.query().first()
    assert tournament.age.lower == 6
    assert tournament.age.upper == 10
    assert tournament.spots.lower == 2
    assert tournament.spots.upper == 11


def test_activity_date_ranges(session, owner, collections):
    sport = collections.activities.add("Sport", username=owner.username)
    police = collections.activities.add("Police", username=owner.username)

    period = collections.periods.add(
        title="Spring 2017",
        prebooking=(datetime(2017, 2, 1), datetime(2017, 2, 28)),
        execution=(datetime(2017, 3, 1), datetime(2017, 3, 31)),
        active=True
    )

    collections.occasions.add(
        start=datetime(2017, 3, 1, 10),
        end=datetime(2017, 3, 1, 12),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=sport,
        period=period
    )

    collections.occasions.add(
        start=datetime(2017, 3, 1, 14),
        end=datetime(2017, 3, 1, 16),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=sport,
        period=period
    )

    collections.occasions.add(
        start=datetime(2017, 3, 4, 8),
        end=datetime(2017, 3, 6, 16),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=sport,
        period=period
    )

    collections.occasions.add(
        start=datetime(2017, 3, 6, 18),
        end=datetime(2017, 3, 8, 18),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=police,
        period=period
    )

    transaction.commit()

    a = collections.activities

    # toggle daterange on
    a = a.for_filter(daterange=(date(2017, 3, 1), date(2017, 3, 30)))
    assert a.query().count() == 2

    # toggle daterange off
    a = a.for_filter(daterange=(date(2017, 3, 1), date(2017, 3, 30)))
    assert a.query().count() == 2  # no filter -> show all
    assert not a.filter.dateranges

    # only include the first activity
    a = a.for_filter(daterange=(date(2017, 3, 1), date(2017, 3, 5)))
    assert a.query().count() == 1

    # include the second activity
    a = a.for_filter(daterange=(date(2017, 3, 7), date(2017, 3, 8)))
    assert a.query().count() == 2

    # exlucde the first activity
    a = a.for_filter(daterange=(date(2017, 3, 1), date(2017, 3, 5)))
    assert a.query().count() == 1


def test_activity_weekdays(session, owner, collections):
    sport = collections.activities.add("Sport", username=owner.username)
    police = collections.activities.add("Police", username=owner.username)

    period = collections.periods.add(
        title="Spring 2017",
        prebooking=(datetime(2017, 2, 1), datetime(2017, 2, 28)),
        execution=(datetime(2017, 5, 8), datetime(2017, 5, 21)),
        active=True
    )

    # Monday
    collections.occasions.add(
        start=datetime(2017, 5, 8, 10),
        end=datetime(2017, 5, 8, 12),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=sport,
        period=period
    )

    # Tuesday
    collections.occasions.add(
        start=datetime(2017, 5, 9, 10),
        end=datetime(2017, 5, 9, 12),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=sport,
        period=period
    )

    # Tuesday - Wednesdy
    collections.occasions.add(
        start=datetime(2017, 5, 9, 18),
        end=datetime(2017, 5, 10, 18),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=police,
        period=period
    )

    transaction.commit()

    a = collections.activities

    # monday
    a = a.for_filter(weekday=0)
    assert a.query().count() == 1

    # monday, tuesday
    a = a.for_filter(weekday=1)
    assert a.query().count() == 2

    # tuesday (negates monday)
    a = a.for_filter(weekday=0)
    assert a.query().count() == 2

    # tuesday, wednesday
    a = a.for_filter(weekday=2)
    assert a.query().count() == 2

    # wednesday (negates tuesday)
    a = a.for_filter(weekday=1)
    assert a.query().count() == 1


def test_profiles(session, owner):

    activities = ActivityCollection(session)
    occasions = OccasionCollection(session)
    periods = PeriodCollection(session)

    sport = activities.add("Sport", username=owner.username)

    autumn = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31))
    )

    winter = periods.add(
        title="Winter 2016",
        prebooking=(datetime(2016, 11, 1), datetime(2016, 11, 30)),
        execution=(datetime(2016, 12, 1), datetime(2016, 12, 31))
    )

    autumn_id, winter_id = autumn.id, winter.id

    # 2 occasions in autumn
    occasions.add(
        start=datetime(2016, 10, 1, 13),
        end=datetime(2016, 10, 1, 14),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=sport,
        period=autumn
    )

    occasions.add(
        start=datetime(2016, 10, 1, 13),
        end=datetime(2016, 10, 1, 14),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=sport,
        period=autumn
    )

    # 1 occasion in winter
    occasions.add(
        start=datetime(2016, 12, 1, 13),
        end=datetime(2016, 12, 1, 14),
        timezone="Europe/Zurich",
        age=(6, 9),
        spots=(2, 10),
        activity=sport,
        period=winter
    )

    transaction.commit()

    sport = activities.query().first()
    period_ids = set(o.period_id for o in sport.occasions)

    assert len(period_ids) == 2
    assert winter_id in period_ids
    assert autumn_id in period_ids

    # drop the winter occasion
    occasions.delete(
        occasions.query().filter(Occasion.period_id == winter_id).first())
    transaction.commit()

    sport = activities.query().first()
    period_ids = set(o.period_id for o in sport.occasions)

    assert len(period_ids) == 1
    assert autumn_id in period_ids


def test_occasion_daterange_constraint(session, owner):
    period = PeriodCollection(session).add(
        title="Autumn 2016",
        prebooking=(datetime(2000, 1, 1), datetime(2000, 1, 1)),
        execution=(datetime(2001, 1, 1), datetime(2050, 12, 31))
    )

    sport = ActivityCollection(session).add("Sport", username=owner.username)
    occasions = OccasionCollection(session)

    with pytest.raises(sqlalchemy.exc.IntegrityError):
        occasions.add(
            activity=sport,
            start=datetime(2020, 10, 10),
            end=datetime(2010, 10, 10),
            timezone='Europe/Zurich',
            period=period
        )


def test_no_orphan_bookings(session, owner):

    activities = ActivityCollection(session)
    attendees = AttendeeCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)
    bookings = BookingCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True
    )

    tournament = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=activities.add("Sport", username=owner.username),
        period=period
    )

    dustin = attendees.add(
        user=owner,
        name="Dustin Henderson",
        birth_date=date(2002, 9, 8),
        gender='male'
    )

    bookings.add(owner, dustin, tournament)

    transaction.commit()

    with pytest.raises(sqlalchemy.exc.IntegrityError):
        occasions.delete(tournament)


def test_no_orphan_occasions(session, owner):

    activities = ActivityCollection(session)
    attendees = AttendeeCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)
    bookings = BookingCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True
    )

    tournament = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=activities.add("Sport", username=owner.username),
        period=period
    )

    dustin = attendees.add(
        user=owner,
        name="Dustin Henderson",
        birth_date=date(2002, 9, 8),
        gender='male'
    )

    bookings.add(owner, dustin, tournament)

    session.flush()

    with pytest.raises(sqlalchemy.exc.IntegrityError):
        activities.delete(activities.query().first())


def test_occasion_duration(session, owner):

    activities = ActivityCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)

    retreat = activities.add("Management Retreat", username=owner.username)

    def durations(activity):
        occasions = session.query(Occasion)\
            .with_entities(Occasion.duration)\
            .filter_by(activity_id=activity.id)

        return sum(set(o.duration for o in occasions))

    assert not DAYS.has(durations(retreat), DAYS.half)
    assert not DAYS.has(durations(retreat), DAYS.full)
    assert not DAYS.has(durations(retreat), DAYS.many)

    periods.add(
        title="2016",
        prebooking=(datetime(2015, 1, 1), datetime(2015, 12, 31)),
        execution=(datetime(2016, 1, 1), datetime(2016, 12, 31)),
        active=True
    )

    # add an occasion that last half a day
    monday = occasions.add(
        start=datetime(2016, 10, 3, 12),
        end=datetime(2016, 10, 3, 16),
        timezone="Europe/Zurich",
        activity=retreat,
        period=periods.active()
    )
    transaction.commit()

    retreat = activities.query().first()
    assert DAYS.has(durations(retreat), DAYS.half)
    assert not DAYS.has(durations(retreat), DAYS.full)
    assert not DAYS.has(durations(retreat), DAYS.many)

    # add another half a day occasion
    tuesday = occasions.add(
        start=datetime(2016, 10, 4, 12),
        end=datetime(2016, 10, 4, 16),
        timezone="Europe/Zurich",
        activity=retreat,
        period=periods.active()
    )
    transaction.commit()

    retreat = activities.query().first()
    assert DAYS.has(durations(retreat), DAYS.half)
    assert not DAYS.has(durations(retreat), DAYS.full)
    assert not DAYS.has(durations(retreat), DAYS.many)

    # add a full day occasion
    wednesday = occasions.add(
        start=datetime(2016, 10, 5, 8),
        end=datetime(2016, 10, 5, 16),
        timezone="Europe/Zurich",
        activity=retreat,
        period=periods.active()
    )
    transaction.commit()

    retreat = activities.query().first()
    assert DAYS.has(durations(retreat), DAYS.half)
    assert DAYS.has(durations(retreat), DAYS.full)
    assert not DAYS.has(durations(retreat), DAYS.many)

    # add a multi day occasion
    weekend = occasions.add(
        start=datetime(2016, 10, 8, 8),
        end=datetime(2016, 10, 9, 16),
        timezone="Europe/Zurich",
        activity=retreat,
        period=periods.active()
    )
    transaction.commit()

    retreat = activities.query().first()
    assert DAYS.has(durations(retreat), DAYS.half)
    assert DAYS.has(durations(retreat), DAYS.full)
    assert DAYS.has(durations(retreat), DAYS.many)

    # remove a the first half-day occasion (nothing should change)
    occasions.delete(monday)
    transaction.commit()

    retreat = activities.query().first()
    assert DAYS.has(durations(retreat), DAYS.half)
    assert DAYS.has(durations(retreat), DAYS.full)
    assert DAYS.has(durations(retreat), DAYS.many)

    # remove the second half-day occasion (no more half-days)
    occasions.delete(tuesday)
    transaction.commit()

    retreat = activities.query().first()
    assert not DAYS.has(durations(retreat), DAYS.half)
    assert DAYS.has(durations(retreat), DAYS.full)
    assert DAYS.has(durations(retreat), DAYS.many)

    # remove the full day occasion
    occasions.delete(wednesday)
    transaction.commit()

    retreat = activities.query().first()
    assert not DAYS.has(durations(retreat), DAYS.half)
    assert not DAYS.has(durations(retreat), DAYS.full)
    assert DAYS.has(durations(retreat), DAYS.many)

    # remove the remaining occasion
    occasions.delete(weekend)
    transaction.commit()

    retreat = activities.query().first()
    assert not DAYS.has(durations(retreat), DAYS.half)
    assert not DAYS.has(durations(retreat), DAYS.full)
    assert not DAYS.has(durations(retreat), DAYS.many)

    # add an occasion with overnight stay which should be categorised as 'many'
    weekend = occasions.add(
        start=datetime(2018, 10, 1, 13, 30),
        end=datetime(2018, 10, 2, 10),
        timezone="Europe/Zurich",
        activity=retreat,
        period=periods.active()
    )
    transaction.commit()

    retreat = activities.query().first()
    assert not DAYS.has(durations(retreat), DAYS.half)
    assert not DAYS.has(durations(retreat), DAYS.full)
    assert DAYS.has(durations(retreat), DAYS.many)


def test_occasion_duration_with_multiple_dates(collections, owner):
    period = collections.periods.add(
        "Summer 2018", (
            datetime(2018, 5, 1),
            datetime(2018, 5, 31),
        ), (
            datetime(2018, 6, 1),
            datetime(2018, 6, 30)
        )
    )

    def with_dates(*ranges):
        a = collections.activities.add(uuid4().hex, username=owner.username)

        first, *rest = ranges

        occasion = collections.occasions.add(
            start=datetime(*first[0]),
            end=datetime(*first[1]),
            timezone='Europe/Zurich',
            activity=a,
            period=period
        )

        for s, e in rest:
            s, e = datetime(*s), datetime(*e)
            collections.occasions.add_date(occasion, s, e, 'Europe/Zurich')

        return occasion

    occasion = with_dates(
        ((2018, 6, 1, 6, 0), (2018, 6, 1, 7, 0)),
        ((2018, 6, 1, 8, 0), (2018, 6, 1, 9, 0))
    )

    assert occasion.duration == DAYS.half

    occasion = with_dates(
        ((2018, 6, 1, 10, 0), (2018, 6, 1, 12, 0)),
        ((2018, 6, 1, 14, 0), (2018, 6, 1, 16, 0))
    )

    assert occasion.duration == DAYS.half

    occasion = with_dates(
        ((2018, 6, 1, 10, 0), (2018, 6, 1, 12, 0)),
        ((2018, 6, 1, 14, 0), (2018, 6, 1, 17, 0))
    )

    assert occasion.duration == DAYS.full

    occasion = with_dates(
        ((2018, 6, 1, 10, 0), (2018, 6, 1, 12, 0)),
        ((2018, 6, 1, 22, 0), (2018, 6, 2, 2, 0))
    )

    assert occasion.duration == DAYS.full

    occasion = with_dates(
        ((2018, 6, 1, 10, 0), (2018, 6, 1, 12, 0)),
        ((2018, 6, 2, 14, 0), (2018, 6, 2, 16, 0))
    )

    assert occasion.duration == DAYS.many


def test_occasion_durations_query(session, owner):

    activities = ActivityCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)

    retreat = activities.add("Management Retreat", username=owner.username)
    meeting = activities.add("Management Meeting", username=owner.username)

    periods.add(
        title="2016",
        prebooking=(datetime(2015, 1, 1), datetime(2015, 12, 31)),
        execution=(datetime(2016, 1, 1), datetime(2016, 12, 31)),
        active=True
    )

    # the retreat lasts a weekend
    occasions.add(
        start=datetime(2016, 10, 8, 8),
        end=datetime(2016, 10, 9, 16),
        timezone="Europe/Zurich",
        activity=retreat,
        period=periods.active()
    )

    # the meeting has a day and a half-day occasion
    occasions.add(
        start=datetime(2016, 10, 10, 8),
        end=datetime(2016, 10, 10, 12),
        timezone="Europe/Zurich",
        activity=meeting,
        period=periods.active()
    )

    occasions.add(
        start=datetime(2016, 10, 11, 8),
        end=datetime(2016, 10, 11, 16),
        timezone="Europe/Zurich",
        activity=meeting,
        period=periods.active()
    )

    transaction.commit()

    assert activities.query().count() == 2
    assert activities.for_filter(duration=DAYS.half).query().count() == 1
    assert activities.for_filter(duration=DAYS.full).query().count() == 1
    assert activities.for_filter(duration=DAYS.many).query().count() == 1

    assert activities\
        .for_filter(duration=DAYS.half)\
        .for_filter(duration=DAYS.full)\
        .query().count() == 1

    assert activities\
        .for_filter(duration=DAYS.half)\
        .for_filter(duration=DAYS.full)\
        .for_filter(duration=DAYS.many)\
        .query().count() == 2

    assert activities\
        .for_filter(duration=DAYS.half)\
        .for_filter(duration=DAYS.many)\
        .query().count() == 2

    # add an occasion with overnight stay which should be categorised as 'many'
    assert activities.for_filter(duration=DAYS.half).query().count() == 1
    assert activities.for_filter(duration=DAYS.full).query().count() == 1
    assert activities.for_filter(duration=DAYS.many).query().count() == 1

    occasions.add(
        start=datetime(2018, 10, 1, 13, 30),
        end=datetime(2018, 10, 2, 10),
        timezone="Europe/Zurich",
        activity=meeting,
        period=periods.active()
    )
    transaction.commit()

    assert activities.for_filter(duration=DAYS.half).query().count() == 1
    assert activities.for_filter(duration=DAYS.full).query().count() == 1
    assert activities.for_filter(duration=DAYS.many).query().count() == 2


def test_occasion_ages(session, owner):

    activities = ActivityCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)

    retreat = activities.add("Management Retreat", username=owner.username)
    meeting = activities.add("Management Meeting", username=owner.username)

    periods.add(
        title="2016",
        prebooking=(datetime(2015, 1, 1), datetime(2015, 12, 31)),
        execution=(datetime(2016, 1, 1), datetime(2016, 12, 31)),
        active=True
    )

    occasions.add(
        start=datetime(2017, 2, 18, 8),
        end=datetime(2017, 2, 18, 17),
        timezone="Europe/Zurich",
        age=(1, 10),
        activity=retreat,
        period=periods.active()
    )
    occasions.add(
        start=datetime(2018, 2, 19, 8),
        end=datetime(2018, 2, 19, 17),
        timezone="Europe/Zurich",
        age=(10, 20),
        activity=meeting,
        period=periods.active()
    )

    transaction.commit()

    assert activities.for_filter(age_range=(1, 20)).query().count() == 2
    assert activities.for_filter(age_range=(1, 10)).query().count() == 2
    assert activities.for_filter(age_range=(10, 20)).query().count() == 2
    assert activities.for_filter(age_range=(1, 5)).query().count() == 1
    assert activities.for_filter(age_range=(15, 20)).query().count() == 1
    assert activities.for_filter(age_range=(20, 21)).query().count() == 1
    assert activities.for_filter(age_range=(21, 22)).query().count() == 0
    assert activities.for_filter(age_range=(4, 4)).query().count() == 1
    assert activities.for_filter(age_range=(15, 15)).query().count() == 1
    assert activities.for_filter(age_range=(0, 0)).query().count() == 0
    assert activities.for_filter(age_range=(1, 5))\
        .for_filter(age_range=(5, 15)).query().count() == 2

    assert activities.for_filter(age_range=(1, 1))\
        .filter.contains_age_range((0, 2))
    assert activities.for_filter(age_range=(1, 1))\
        .filter.contains_age_range((1, 1))
    assert activities.for_filter(age_range=(1, 1))\
        .filter.contains_age_range((1, 2))

    assert not activities.for_filter(age_range=(1, 1))\
        .filter.contains_age_range((0, 0))

    assert not activities.for_filter(age_range=(1, 10))\
        .filter.contains_age_range((20, 30))

    assert activities\
        .for_filter(age_range=(1, 10))\
        .for_filter(age_range=(20, 30))\
        .filter.contains_age_range((10, 15))

    assert activities\
        .for_filter(age_range=(1, 10))\
        .for_filter(age_range=(20, 30))\
        .filter.contains_age_range((10, 20))

    assert not activities\
        .for_filter(age_range=(1, 10))\
        .for_filter(age_range=(20, 30))\
        .filter.contains_age_range((15, 16))


def test_occasion_owners(session, owner, secondary_owner):

    activities = ActivityCollection(session)

    activities.add("Management Retreat", username=owner.username)
    activities.add("Management Meeting", username=secondary_owner.username)

    transaction.commit()

    assert activities\
        .for_filter(owner='owner@example.org')\
        .query().count() == 1
    assert activities\
        .for_filter(owner='owner@example.org')\
        .for_filter(owner='secondary@example.org')\
        .query().count() == 2

    assert activities\
        .for_filter(owner='secondary@example.org')\
        .query().count() == 1


def test_attendee_age(session, owner):

    def age(years):
        return date.today().replace(year=date.today().year - years)

    attendees = AttendeeCollection(session)
    d = attendees.add(owner, "Dustin Henderson", age(13), 'male')
    m = attendees.add(owner, "Mike Wheeler", age(14), 'male')

    assert d.age == 13
    assert m.age == 14

    assert attendees.query().filter(Attendee.age <= 13).count() == 1
    assert attendees.query().filter(Attendee.age <= 14).count() == 2


def test_booking_collection(session, owner):

    activities = ActivityCollection(session)
    attendees = AttendeeCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)
    bookings = BookingCollection(session)

    tournament = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=activities.add("Sport", username=owner.username),
        period=periods.add(
            title="Autumn 2016",
            prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
            execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
            active=True
        )
    )

    dustin = attendees.add(
        user=owner,
        name="Dustin Henderson",
        birth_date=date(2002, 9, 8),
        gender='male'
    )

    bookings.add(owner, dustin, tournament)

    assert bookings.query().count() == 1
    assert bookings.for_period(Bunch(id=uuid4())).query().count() == 0
    assert bookings.for_username('foobar').query().count() == 0
    assert bookings.for_period(Bunch(id=uuid4())).count(owner.username) == 0
    assert bookings.booking_count(owner.username) == 1


def test_star_nobble_booking(session, owner):
    activities = ActivityCollection(session)
    attendees = AttendeeCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)
    bookings = BookingCollection(session)

    sport = activities.add("Sport", username=owner.username)

    autumn = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True
    )

    s1 = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=sport,
        period=autumn
    )

    s2 = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=sport,
        period=autumn
    )

    s3 = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=sport,
        period=autumn
    )

    dustin = attendees.add(
        user=owner,
        name="Dustin Henderson",
        birth_date=date(2002, 9, 8),
        gender='male'
    )

    b1 = bookings.add(owner, dustin, s1)
    b2 = bookings.add(owner, dustin, s2)
    b3 = bookings.add(owner, dustin, s3)

    assert b1.star(max_stars=1) is True
    assert b2.star(max_stars=1) is False

    assert b1.starred is True
    assert b2.starred is False

    b1.nobble()
    assert b1.starred is True
    assert b1.nobbled is True

    b1.unstar()
    assert b1.starred is False
    assert b1.nobbled is True

    b1.unnobble()
    assert b1.starred is False
    assert b1.nobbled is False

    b1.nobble()
    b1.star()

    b2.unstar()
    b2.unnobble()

    b3.nobble()
    b3.unstar()

    q = session.query(Booking)
    assert q.filter(Booking.starred == True).count() == 1
    assert q.filter(Booking.nobbled == True).count() == 2

    b2.nobble()
    b2.star()

    assert q.filter(Booking.starred == True).count() == 2
    assert q.filter(Booking.nobbled == True).count() == 3

    b1.unstar()
    b1.unnobble()

    assert b1.priority == 0

    b1.star()

    assert b1.priority == 1

    b1.nobble()

    assert b1.priority == 3

    b1.unstar()

    assert b1.priority == 2


def test_booking_period_id_reference(session, owner):

    activities = ActivityCollection(session)
    attendees = AttendeeCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)
    bookings = BookingCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True
    )

    tournament = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=activities.add("Sport", username=owner.username),
        period=period
    )

    dustin = attendees.add(
        user=owner,
        name="Dustin Henderson",
        birth_date=date(2002, 9, 8),
        gender='male'
    )

    bookings.add(owner, dustin, tournament)

    transaction.commit()

    assert bookings.query().first().period_id == period.id

    new = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=False
    )

    tournament = occasions.query().first()
    tournament.period_id = new.id

    transaction.commit()

    assert bookings.query().first().period_id == new.id


def test_happiness(session, owner):

    activities = ActivityCollection(session)
    attendees = AttendeeCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)
    bookings = BookingCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True
    )

    sport = activities.add("Sport", username=owner.username)

    o1 = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=sport,
        period=period
    )

    o2 = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=sport,
        period=period
    )

    o3 = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=sport,
        period=period
    )

    dustin = attendees.add(
        user=owner,
        name="Dustin Henderson",
        birth_date=date(2002, 9, 8),
        gender='male'
    )

    transaction.commit()

    def assert_happiness(period_id, value):

        def equal(result):
            if value is None:
                return result is None
            else:
                return round(result, 3) == value

        dustin = attendees.query().first()
        assert equal(dustin.happiness(period_id))

        q = attendees.query().with_entities(Attendee.happiness(period_id))
        assert equal(q.first().happiness)

    # no bookings yet
    assert_happiness(period.id, None)

    bookings.add(owner, dustin, o1)
    bookings.add(owner, dustin, o2)
    bookings.add(owner, dustin, o3)

    transaction.commit()

    # low priorities, all bookings accepted
    b1, b2, b3 = bookings.query().all()
    b1.state = 'accepted'
    b2.state = 'accepted'
    b3.state = 'accepted'

    transaction.commit()

    assert_happiness(period.id, 1.0)

    # high priorities, all bookings accepted
    b1, b2, b3 = bookings.query().all()
    b1.priority = 1
    b2.priority = 1
    b3.priority = 1

    transaction.commit()

    assert_happiness(period.id, 1.0)

    # low priorities denied, one high priority accepted
    b1, b2, b3 = bookings.query().all()
    b1.priority = 1
    b1.state = 'accepted'
    b2.priority = 0
    b2.state = 'open'
    b3.priority = 0
    b3.state = 'open'

    transaction.commit()

    assert_happiness(period.id, 0.5)

    # low priorities accepted, one high priority denied
    b1, b2, b3 = bookings.query().all()
    b1.priority = 1
    b1.state = 'open'
    b2.priority = 0
    b2.state = 'accepted'
    b3.priority = 0
    b3.state = 'accepted'

    transaction.commit()

    assert_happiness(period.id, 0.5)

    # 1/3 high priorities accepted
    b1, b2, b3 = bookings.query().all()
    b1.priority = 1
    b1.state = 'accepted'
    b2.priority = 1
    b2.state = 'open'
    b3.priority = 1
    b3.state = 'open'

    transaction.commit()

    assert_happiness(period.id, 0.333)

    # 1 extra high priority booking accepted
    b1, b2, b3 = bookings.query().all()
    b1.priority = 7
    b1.state = 'accepted'
    b2.priority = 0
    b2.state = 'open'
    b3.priority = 0
    b3.state = 'open'

    transaction.commit()

    assert_happiness(period.id, 0.8)


def test_attendees_count(session, owner):
    activities = ActivityCollection(session)
    attendees = AttendeeCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)
    bookings = BookingCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True
    )

    sport = activities.add("Sport", username=owner.username)

    o = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=sport,
        period=period,
        spots=(1, 2)
    )

    a1 = attendees.add(
        user=owner,
        name="Dustin Henderson",
        birth_date=date(2002, 9, 8),
        gender='male'
    )

    a2 = attendees.add(
        user=owner,
        name="Mike Wheeler",
        birth_date=date(2002, 8, 8),
        gender='male'
    )

    transaction.commit()

    q = occasions.query()

    assert q.one().attendee_count == 0
    assert q.one().available_spots == 2
    assert q.with_entities(Occasion.available_spots).one()[0] == 2

    assert not q.one().operable

    bookings.add(owner, a1, o)
    transaction.commit()

    assert q.one().attendee_count == 0
    assert q.one().available_spots == 2
    assert q.with_entities(Occasion.available_spots).one()[0] == 2
    assert not q.one().operable

    bookings.query().one().state = 'accepted'
    transaction.commit()

    assert q.one().attendee_count == 1
    assert q.one().available_spots == 1
    assert q.with_entities(Occasion.available_spots).one()[0] == 1
    assert q.one().operable

    bookings.add(owner, a2, o)
    transaction.commit()

    assert q.one().attendee_count == 1
    assert q.one().available_spots == 1
    assert q.with_entities(Occasion.available_spots).one()[0] == 1
    assert q.one().operable

    bookings.query().all()[0].state = 'accepted'
    bookings.query().all()[1].state = 'accepted'
    transaction.commit()

    assert q.one().attendee_count == 2
    assert q.one().available_spots == 0
    assert q.with_entities(Occasion.available_spots).one()[0] == 0
    assert q.one().operable

    session.delete(bookings.query().first())
    transaction.commit()

    assert q.one().attendee_count == 1
    assert q.one().available_spots == 1
    assert q.with_entities(Occasion.available_spots).one()[0] == 1
    assert q.one().operable

    bookings.query().one().state = 'open'

    assert q.one().attendee_count == 0
    assert q.one().available_spots == 2
    assert q.with_entities(Occasion.available_spots).one()[0] == 2
    assert not q.one().operable

    q.one().cancelled = True
    transaction.commit()

    assert q.one().attendee_count == 0
    assert q.one().available_spots == 0
    assert q.with_entities(Occasion.available_spots).one()[0] == 0


def test_accept_booking(session, owner):
    activities = ActivityCollection(session)
    attendees = AttendeeCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)
    bookings = BookingCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True,
    )

    o1 = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=activities.add("Activity 1", username=owner.username),
        period=period,
        spots=(0, 2)
    )

    o2 = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=activities.add("Activity 2", username=owner.username),
        period=period,
        spots=(0, 2)
    )

    o3 = occasions.add(
        start=datetime(2016, 10, 5, 13),
        end=datetime(2016, 10, 5, 14),
        timezone="Europe/Zurich",
        activity=activities.add("Activity 3", username=owner.username),
        period=period,
        spots=(0, 2)
    )

    o4 = occasions.add(
        start=datetime(2016, 10, 5, 15),
        end=datetime(2016, 10, 5, 16),
        timezone="Europe/Zurich",
        activity=activities.add("Activity 4", username=owner.username),
        period=period,
        spots=(0, 2)
    )

    a1 = attendees.add(
        user=owner,
        name="Dustin Henderson",
        birth_date=date(2000, 1, 1),
        gender='male'
    )

    a2 = attendees.add(
        user=owner,
        name="Mike Wheeler",
        birth_date=date(2000, 1, 1),
        gender='male'
    )

    a3 = attendees.add(
        user=owner,
        name="Eleven",
        birth_date=date(2000, 1, 1),
        gender='female'
    )

    transaction.commit()

    # only works for confirmed periods
    with pytest.raises(RuntimeError) as e:
        bookings.accept_booking(bookings.add(owner, a1, o1))

    assert "The period has not yet been confirmed" in str(e)
    transaction.abort()

    periods.active().confirmed = True
    transaction.commit()

    # adding bookings works until the occasion is full
    bookings.accept_booking(bookings.add(owner, a1, o1))
    bookings.accept_booking(bookings.add(owner, a2, o1))

    with pytest.raises(RuntimeError) as e:
        bookings.accept_booking(bookings.add(owner, a3, o1))

    assert "The occasion is already full" in str(e)
    transaction.abort()

    # only open/denied bookings can be accepted this way
    booking = bookings.add(owner, a1, o1)

    for state in ('accepted', 'blocked', 'cancelled'):
        booking.state = state

        with pytest.raises(RuntimeError) as e:
            bookings.accept_booking(booking)

    transaction.abort()

    # we can't accept a booking that conflicts with another accepted booking
    bookings.accept_booking(bookings.add(owner, a1, o1))

    with pytest.raises(RuntimeError) as e:
        bookings.accept_booking(bookings.add(owner, a1, o2))

    assert "Conflict with booking" in str(e)
    transaction.abort()

    # other conflicting bookings are marked as blocked
    b1 = bookings.add(owner, a1, o1)
    b2 = bookings.add(owner, a1, o2)

    bookings.accept_booking(b1)

    assert b1.state == 'accepted'
    assert b2.state == 'blocked'

    b1.state = 'open'
    b2.state = 'open'

    bookings.accept_booking(b2)

    assert b2.state == 'accepted'
    assert b1.state == 'blocked'

    transaction.abort()

    # if there's a booking limit we ensure it isn't violated
    period = periods.active()
    period.all_inclusive = True
    period.max_bookings_per_attendee = 1

    b1 = bookings.add(owner, a1, o1)
    bookings.accept_booking(b1)

    with pytest.raises(RuntimeError) as e:
        bookings.accept_booking(bookings.add(owner, a1, o3))

    assert "The booking limit has been reached" in str(e)

    transaction.abort()

    # there was a regression which lead to an error before the booking limit
    # was reached (accepted bookings were assumed to be in conflict)
    period = periods.active()
    period.all_inclusive = True
    period.max_bookings_per_attendee = 2

    b1 = bookings.add(owner, a1, o1)
    bookings.accept_booking(b1)
    b2 = bookings.add(owner, a1, o3)
    bookings.accept_booking(b2)

    with pytest.raises(RuntimeError) as e:
        bookings.accept_booking(bookings.add(owner, a1, o4))

    assert "The booking limit has been reached" in str(e)

    transaction.abort()

    # the custom limit overrides the period limit
    a1 = attendees.query().filter(Attendee.name == "Dustin Henderson").one()
    a1.limit = 1

    period = periods.active()
    period.max_bookings_per_attendee = 2

    bookings.accept_booking(bookings.add(owner, a1, o1))

    with pytest.raises(RuntimeError) as e:
        bookings.accept_booking(bookings.add(owner, a1, o3))

    assert "The booking limit has been reached" in str(e)

    transaction.abort()

    # if accepting the booking leads to the booking limit, the rest is blocked
    period = periods.active()
    period.all_inclusive = True
    period.max_bookings_per_attendee = 1

    assert o1.dates[0].end < o3.dates[0].start
    b1 = bookings.add(owner, a1, o1)
    b2 = bookings.add(owner, a1, o3)

    bookings.accept_booking(b1)

    assert b1.state == 'accepted'
    assert b2.state == 'blocked'


def test_booking_limit_exemption(session, owner):
    activities = ActivityCollection(session)
    attendees = AttendeeCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)
    bookings = BookingCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True,
    )
    period.confirmed = True

    o1 = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=activities.add("Activity 1", username=owner.username),
        period=period,
        spots=(0, 2)
    )
    o1.exempt_from_booking_limit = True

    o2 = occasions.add(
        start=datetime(2016, 11, 4, 13),
        end=datetime(2016, 11, 4, 14),
        timezone="Europe/Zurich",
        activity=activities.add("Activity 2", username=owner.username),
        period=period,
        spots=(0, 2)
    )
    o2.exempt_from_booking_limit = False

    a1 = attendees.add(
        user=owner,
        name="Dustin Henderson",
        birth_date=date(2000, 1, 1),
        gender='male'
    )

    a2 = attendees.add(
        user=owner,
        name="Mike Wheeler",
        birth_date=date(2000, 1, 1),
        gender='male'
    )

    transaction.commit()

    # 1st case, get the exempt booking, then the non-exempt one
    b1 = bookings.add(owner, a1, o1)
    b2 = bookings.add(owner, a1, o2)
    b1.state = 'open'
    b2.state = 'open'

    bookings.accept_booking(b1)
    bookings.accept_booking(b2)

    # 2nd case, get the non-exempt booking, then the exempt one
    b3 = bookings.add(owner, a2, o1)
    b4 = bookings.add(owner, a2, o2)
    b3.state = 'open'
    b4.state = 'open'

    bookings.accept_booking(b3)
    bookings.accept_booking(b4)

    # none of these methods should throw an error, then we're good


def test_cancel_booking(session, owner):
    activities = ActivityCollection(session)
    attendees = AttendeeCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)
    bookings = BookingCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True,
    )

    o1 = occasions.add(
        start=datetime(2016, 10, 4, 9),
        end=datetime(2016, 10, 4, 12),
        timezone="Europe/Zurich",
        activity=activities.add("Activity 1", username=owner.username),
        period=period,
        spots=(0, 2)
    )

    o2 = occasions.add(
        start=datetime(2016, 10, 4, 11),
        end=datetime(2016, 10, 4, 14),
        timezone="Europe/Zurich",
        activity=activities.add("Activity 2", username=owner.username),
        period=period,
        spots=(0, 2)
    )

    o3 = occasions.add(
        start=datetime(2016, 10, 4, 13),
        end=datetime(2016, 10, 4, 16),
        timezone="Europe/Zurich",
        activity=activities.add("Activity 3", username=owner.username),
        period=period,
        spots=(0, 2)
    )

    o4 = occasions.add(
        start=datetime(2016, 10, 4, 15),
        end=datetime(2016, 10, 4, 18),
        timezone="Europe/Zurich",
        activity=activities.add("Activity 4", username=owner.username),
        period=period,
        spots=(0, 1)
    )

    a1 = attendees.add(
        user=owner,
        name="Dustin Henderson",
        birth_date=date(2000, 1, 1),
        gender='male'
    )

    a2 = attendees.add(
        user=owner,
        name="Mike Wheeler",
        birth_date=date(2000, 1, 1),
        gender='male'
    )

    a3 = attendees.add(
        user=owner,
        name="Eleven",
        birth_date=date(2000, 1, 1),
        gender='female'
    )

    transaction.commit()

    # only works for confirmed periods
    with pytest.raises(RuntimeError) as e:
        bookings.cancel_booking(bookings.add(owner, a1, o1))

    assert "The period has not yet been confirmed" in str(e)
    transaction.abort()

    periods.active().confirmed = True
    transaction.commit()

    # cancelling a booking will automatically accept the blocked ones
    # (this is run after matching, so we want to make sure the matching
    # is kept tight, with no unnecessarily open/denied bookings)
    b1 = bookings.add(owner, a1, o1)
    b2 = bookings.add(owner, a1, o2)
    b3 = bookings.add(owner, a1, o3)

    bookings.accept_booking(b2)

    assert b1.state == 'blocked'
    assert b2.state == 'accepted'
    assert b1.state == 'blocked'

    bookings.cancel_booking(b2)

    assert b1.state == 'accepted'
    assert b2.state == 'cancelled'
    assert b3.state == 'accepted'

    transaction.abort()

    # same, this time with only one overlap
    b1 = bookings.add(owner, a1, o1)
    b2 = bookings.add(owner, a1, o2)
    b3 = bookings.add(owner, a1, o3)

    bookings.accept_booking(b1)

    assert b1.state == 'accepted'
    assert b2.state == 'blocked'
    assert b3.state == 'open'

    bookings.cancel_booking(b1)

    assert b1.state == 'cancelled'
    assert b2.state == 'accepted'
    assert b3.state == 'blocked'

    transaction.abort()

    # if the occasions are already full, the state is going to be 'denied'
    bookings.accept_booking(bookings.add(owner, a2, o1))
    bookings.accept_booking(bookings.add(owner, a3, o1))

    b1 = bookings.add(owner, a1, o1)
    b2 = bookings.add(owner, a1, o2)

    bookings.accept_booking(b2)

    assert b1.state == 'blocked'
    assert b2.state == 'accepted'

    bookings.cancel_booking(b2)

    assert b1.state == 'denied'
    assert b2.state == 'cancelled'

    transaction.abort()

    # if the cancellation leads to open spots, other bookings are considered
    b1 = bookings.add(owner, a1, o1)
    b2 = bookings.add(owner, a2, o1)
    b3 = bookings.add(owner, a3, o1)

    bookings.accept_booking(b1)
    bookings.accept_booking(b2)

    assert b1.state == 'accepted'
    assert b2.state == 'accepted'
    assert b3.state == 'open'

    bookings.cancel_booking(b2)

    assert b1.state == 'accepted'
    assert b2.state == 'cancelled'
    assert b3.state == 'accepted'

    transaction.abort()

    # make sure a cancellation doesn't lead to overbooking
    b1 = bookings.add(owner, a1, o4)
    b2 = bookings.add(owner, a2, o4, priority=1)
    b3 = bookings.add(owner, a3, o4)

    bookings.accept_booking(b1)

    assert b1.state == 'accepted'
    assert b2.state == 'open'
    assert b3.state == 'open'

    bookings.cancel_booking(b1)

    assert b1.state == 'cancelled'
    assert b2.state == 'accepted'
    assert b3.state == 'open'

    transaction.abort()

    # make sure the booking limit is honored
    period = periods.active()
    period.all_inclusive = True
    period.max_bookings_per_attendee = 1

    b1 = bookings.add(owner, a1, o1, priority=4)
    b2 = bookings.add(owner, a1, o2, priority=3)
    b3 = bookings.add(owner, a1, o3, priority=2)
    b4 = bookings.add(owner, a1, o4, priority=1)

    bookings.accept_booking(b1)

    assert b1.state == 'accepted'
    assert b2.state == 'blocked'
    assert b3.state == 'blocked'
    assert b4.state == 'blocked'

    bookings.cancel_booking(b1)

    assert b1.state == 'cancelled'
    assert b2.state == 'accepted'
    assert b3.state == 'blocked'
    assert b4.state == 'blocked'

    bookings.cancel_booking(b2)

    assert b1.state == 'cancelled'
    assert b2.state == 'cancelled'
    assert b3.state == 'accepted'
    assert b4.state == 'blocked'

    transaction.abort()

    # make sure accepting a previously denied booking of the same occasion will
    # will be skipped if doing so would conflict with the limit
    period = periods.active()
    period.all_inclusive = True
    period.max_bookings_per_attendee = 1

    b1 = bookings.add(owner, a1, o4)
    b2 = bookings.add(owner, a2, o1)
    b3 = bookings.add(owner, a2, o4)

    b1.state = 'accepted'
    b2.state = 'accepted'
    b3.state = 'denied'

    session.flush()

    bookings.cancel_booking(b1)

    assert b1.state == 'cancelled'
    assert b2.state == 'accepted'
    assert b3.state == 'denied'


def test_period_phases(session):
    periods = PeriodCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(date(2016, 9, 1), date(2016, 9, 30)),
        execution=(date(2016, 11, 1), date(2016, 11, 30)),
        active=False,
    )

    assert period.phase == 'inactive'

    period.active = True

    with freeze_time('2016-08-31'):
        assert period.phase == 'inactive'

    with freeze_time('2016-09-01'):
        assert period.phase == 'wishlist'

    period.confirmed = True
    assert period.phase == 'booking'

    period.finalized = True

    with freeze_time('2016-10-31'):
        assert period.phase == 'payment'

    with freeze_time('2016-11-01'):
        assert period.phase == 'execution'

    with freeze_time('2016-12-01'):
        assert period.phase == 'archive'


def test_invoices(session, owner, prebooking_period, inactive_period):
    p1 = prebooking_period
    p2 = inactive_period

    invoices = InvoiceCollection(session, user_id=owner.id)

    assert invoices.total_amount == 0
    assert invoices.outstanding_amount == 0
    assert invoices.paid_amount == 0

    i1 = invoices.add(period_id=p1.id)
    i1.add("Malcolm", "Camp", 100.0, 1.0)
    i1.add("Malcolm", "Pass", 25.0, 1.0)
    i1.add("Dewey", "Football", 100.0, 1.0)
    i1.add("Dewey", "Pass", 25.0, 1.0)
    i1.add("Discount", "8%", 250, -0.05)

    assert i1.total_amount == 237.5
    assert i1.outstanding_amount == 237.5
    assert i1.paid_amount == 0
    assert InvoiceCollection(session).total_amount == 237.5

    i2 = invoices.add(period_id=p2.id)
    i2.add("Malcolm", "Camp", 100, 1 / 3)

    # this is 33, not 33.33 because both unit and quantity are truncated
    # to two decimals after the point. So when we added 1 / 3 above, we really
    # added 0.33 to the database (100 * 0.33 = 33)
    assert InvoiceCollection(session, period_id=p2.id).total_amount == 33
    assert i2.total_amount == 33
    assert i2.outstanding_amount == 33
    assert i2.paid_amount == 0

    assert InvoiceCollection(session, period_id=p1.id).total_amount == 237.5
    assert InvoiceCollection(session, period_id=p2.id).total_amount == 33
    assert InvoiceCollection(session).total_amount == 270.5

    # pay part of the first invoice
    i1.items[0].paid = True
    assert i1.total_amount == 237.5
    assert i1.outstanding_amount == 137.5
    assert i1.paid_amount == 100.0
    assert i1.paid == False

    assert session.query(func.sum(Invoice.total_amount))\
        .first()[0] == 270.5
    assert session.query(func.sum(Invoice.outstanding_amount))\
        .first()[0] == 170.5
    assert session.query(func.sum(Invoice.paid_amount))\
        .first()[0] == 100

    assert InvoiceCollection(session, period_id=p1.id)\
        .outstanding_amount == 137.5
    assert InvoiceCollection(session, period_id=p2.id)\
        .outstanding_amount == 33

    assert InvoiceCollection(session).unpaid_count() == 2
    assert InvoiceCollection(session).unpaid_count(
        excluded_period_ids=(p1.id, )) == 1

    # pay all of the second invoice
    i2.items[0].paid = True
    assert i2.total_amount == 33
    assert i2.outstanding_amount == 0
    assert i2.paid_amount == 33
    assert i2.paid == True

    assert InvoiceCollection(session).unpaid_count() == 1


def test_invoice_reference(session, owner, prebooking_period):
    invoices = InvoiceCollection(
        session, user_id=owner.id, period_id=prebooking_period.id)

    # DEFAULT SCHEMA
    # --------------
    i1 = invoices.add()
    assert len(i1.references) == 1
    assert i1.references[0].reference.startswith('q')
    assert i1.references[0].readable.startswith('Q')
    assert i1.references[0].schema == 'feriennet-v1'
    assert i1.references[0].bucket == 'feriennet-v1'

    # make sure linking to the same schema is idempotent
    invoices.schema.link(session, i1)
    session.refresh(i1)
    assert len(i1.references) == 1

    # ESR
    # ---
    invoices = invoices.for_schema('esr-v1')
    invoices.schema.link(session, i1)
    session.refresh(i1)
    assert len(i1.references) == 2
    assert len(i1.references[1].reference) == 27
    assert ' ' in i1.references[1].readable
    assert i1.references[1].schema == 'esr-v1'
    assert i1.references[1].bucket == 'esr-v1'

    # make sure linking to the same schema is idempotent
    invoices.schema.link(session, i1)
    session.refresh(i1)
    assert len(i1.references) == 2

    # Raiffeisen
    # ----------
    invoices = invoices.for_schema('raiffeisen-v1', schema_config=dict(
        esr_identification_number='123123'
    ))
    invoices.schema.link(session, i1)
    session.refresh(i1)
    assert len(i1.references) == 3
    assert len(i1.references[2].reference) == 27
    assert i1.references[2].reference.startswith('123123')
    assert i1.references[2].schema == 'raiffeisen-v1'
    assert i1.references[2].bucket.startswith('raiffeisen-v1-')

    # make sure linking the same schema is idempotent
    invoices.schema.link(session, i1)
    session.refresh(i1)
    assert len(i1.references) == 3

    # unless we change the config
    invoices = invoices.for_schema('raiffeisen-v1', schema_config=dict(
        esr_identification_number='321321'
    ))

    invoices.schema.link(session, i1)
    session.refresh(i1)
    assert len(i1.references) == 4
    assert i1.references[3].reference.startswith('321321')
    assert i1.references[3].schema == 'raiffeisen-v1'
    assert i1.references[3].bucket.startswith('raiffeisen-v1-')
    assert i1.references[2].bucket != i1.references[3].bucket


def test_invoice_reference_format_esr():

    schema = ESRSchema()

    assert schema.format('000127131108141601011502061')\
        == '1271 31108 14160 10115 02061'

    assert schema.format('127131108141601011502061')\
        == '1271 31108 14160 10115 02061'

    assert schema.checksum('96111690000000660000000928') == '4'
    assert schema.checksum('12000000000023447894321689') == '9'


def test_invoice_reference_format_feriennet():
    assert FeriennetSchema().format('qeb3afd0e43') == 'Q-EB3AF-D0E43'


def test_invoice_reference_extract_feriennet_schema():
    extract_code = FeriennetSchema().extract

    assert extract_code('') is None
    assert extract_code('\n asdf') is None
    assert extract_code('Q-70171-292FA') == 'q70171292fa'
    assert extract_code('B-70171-292FA') is None
    assert extract_code('Q-7o171-292FA') == 'q70171292fa'
    assert extract_code('Q-7o171-292FA') == 'q70171292fa'
    assert extract_code('Q-   7o171292FA') == 'q70171292fa'
    assert extract_code('Q-   7o171  29  ---- 2FA') == 'q70171292fa'
    assert extract_code('Q\n7o171\n292FA') == 'q70171292fa'
    assert extract_code('Code: q-70171-292fa') == 'q70171292fa'


def test_invoice_reference_uniqueness(session, owner, prebooking_period):
    invoices = InvoiceCollection(
        session, user_id=owner.id, period_id=prebooking_period.id)

    i1 = invoices.add()

    with pytest.raises(IntegrityError):
        invoices.schema.link(session, i1, optimistic=True)


def test_confirm_period(session, owner):

    activities = ActivityCollection(session)
    attendees = AttendeeCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)
    bookings = BookingCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True)
    period.all_inclusive = False
    period.booking_cost = 10

    sport = activities.add("Sport", username=owner.username)

    o = occasions.add(
        start=datetime(2016, 10, 4, 10),
        end=datetime(2016, 10, 4, 12),
        timezone="Europe/Zurich",
        activity=sport,
        period=period,
        spots=(0, 2))
    o.cost = 20

    a1 = attendees.add(
        user=owner,
        name="Dustin Henderson",
        birth_date=date(2000, 1, 1),
        gender='male')

    a2 = attendees.add(
        user=owner,
        name="Mike Wheeler",
        birth_date=date(2000, 1, 1),
        gender='male')

    transaction.commit()

    b1 = bookings.add(owner, a1, o)
    b2 = bookings.add(owner, a2, o)
    b1.state = 'open'
    b2.state = 'accepted'

    period = periods.query().one()
    period.confirm()

    assert bookings.query().all()[0].cost == 30.0
    assert bookings.query().all()[1].cost == 30.0
    assert sorted([b.state for b in bookings.query()]) == [
        'accepted',
        'denied',
    ]

    transaction.abort()

    period = periods.query().one()
    period.all_inclusive = True
    period.booking_cost = 10

    b1 = bookings.add(owner, a1, o)

    period.confirm()

    assert bookings.query().one().cost == 20.0


def test_cancel_occasion(session, owner):

    activities = ActivityCollection(session)
    attendees = AttendeeCollection(session)
    periods = PeriodCollection(session)
    occasions = OccasionCollection(session)
    bookings = BookingCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True)

    o1 = occasions.add(
        start=datetime(2016, 10, 4, 10),
        end=datetime(2016, 10, 4, 12),
        timezone="Europe/Zurich",
        activity=activities.add("Sport", username=owner.username),
        period=period,
        spots=(0, 2))

    o2 = occasions.add(
        start=datetime(2016, 10, 4, 10),
        end=datetime(2016, 10, 4, 12),
        timezone="Europe/Zurich",
        activity=activities.add("Science", username=owner.username),
        period=period,
        spots=(0, 2))

    a1 = attendees.add(
        user=owner,
        name="Dustin Henderson",
        birth_date=date(2008, 1, 1),
        gender='male')

    a2 = attendees.add(
        user=owner,
        name="Mike Wheeler",
        birth_date=date(2008, 1, 1),
        gender='male')

    transaction.commit()

    o1, o2 = occasions.query().all()

    b1 = bookings.add(owner, a1, o1)
    b2 = bookings.add(owner, a2, o1)

    o1.cancel()
    assert b1.state == 'cancelled'
    assert b2.state == 'cancelled'
    assert o1.cancelled
    assert not o2.cancelled

    transaction.abort()

    periods.active().confirmed = True
    o1, o2 = occasions.query().all()

    b1 = bookings.add(owner, a1, o1)
    b2 = bookings.add(owner, a1, o2)

    b1.state = 'accepted'
    b2.state = 'blocked'

    o1.cancel()

    assert b1.state == 'cancelled'
    assert b2.state == 'accepted'
    assert o1.cancelled
    assert not o2.cancelled


def test_no_overlapping_dates(session, collections, prebooking_period, owner):
    period = prebooking_period

    o = collections.occasions.add(
        start=period.execution_start,
        end=period.execution_start + timedelta(hours=2),
        timezone="Europe/Zurich",
        activity=collections.activities.add("Sport", username=owner.username),
        period=period
    )

    with pytest.raises(AssertionError):
        collections.occasions.add_date(
            occasion=o,
            start=period.execution_start + timedelta(hours=1),
            end=period.execution_start + timedelta(hours=3),
            timezone="Europe/Zurich"
        )


def test_no_occasion_orphans(session, collections, prebooking_period, owner):
    period = prebooking_period

    collections.occasions.add(
        start=period.execution_start,
        end=period.execution_start + timedelta(hours=2),
        timezone="Europe/Zurich",
        activity=collections.activities.add("Sport", username=owner.username),
        period=period
    )

    transaction.commit()

    assert collections.occasions.query().count() == 1
    assert session.query(OccasionDate).count() == 1

    collections.occasions.delete(collections.occasions.query().first())
    transaction.commit()

    assert collections.occasions.query().count() == 0
    assert session.query(OccasionDate).count() == 0


def test_date_changes(session, collections, prebooking_period, owner):
    period = prebooking_period

    collections.occasions.add(
        start=period.execution_start,
        end=period.execution_start + timedelta(hours=2),
        timezone="Europe/Zurich",
        activity=collections.activities.add("Sport", username=owner.username),
        period=period
    )

    transaction.commit()

    o = collections.occasions.query().first()
    assert DAYS.has(o.duration, DAYS.half)
    assert not DAYS.has(o.duration, DAYS.full)
    assert not DAYS.has(o.duration, DAYS.many)

    o.dates[0].end += timedelta(hours=6)
    assert DAYS.has(o.duration, DAYS.half)
    assert not DAYS.has(o.duration, DAYS.full)
    assert not DAYS.has(o.duration, DAYS.many)

    session.flush()

    assert not DAYS.has(o.duration, DAYS.half)
    assert DAYS.has(o.duration, DAYS.full)
    assert not DAYS.has(o.duration, DAYS.many)


def test_is_permitted_birth_date():
    o = Occasion(age=NumericRange(6, 9), dates=[
        OccasionDate(
            start=datetime(2017, 7, 26, 10),
            end=datetime(2017, 7, 26, 16),
            timezone='Europe/Zurich'
        ),
        OccasionDate(
            start=datetime(2017, 7, 26, 17),
            end=datetime(2017, 7, 26, 20),
            timezone='Europe/Zurich'
        )
    ])

    assert o.is_permitted_birth_date(date(2012, 7, 26), wiggle_room=366)
    assert not o.is_permitted_birth_date(date(2012, 7, 26), wiggle_room=365)

    assert not o.is_too_young(date(2012, 7, 26), wiggle_room=366)
    assert o.is_too_young(date(2012, 7, 26), wiggle_room=365)

    assert o.is_permitted_birth_date(date(2008, 7, 26), wiggle_room=365)
    assert not o.is_permitted_birth_date(date(2008, 7, 26), wiggle_room=364)

    assert not o.is_too_old(date(2008, 7, 26), wiggle_room=365)
    assert o.is_too_old(date(2008, 7, 26), wiggle_room=364)


def test_deadline(session, collections, prebooking_period, owner):
    period = prebooking_period

    start, end = period.execution_start,\
        period.execution_start + timedelta(hours=2)

    occasion = collections.occasions.add(
        start=start,
        end=end,
        timezone="Europe/Zurich",
        activity=collections.activities.add("Sport", username=owner.username),
        period=period
    )
    assert occasion.deadline == period.execution_start.date()

    period.deadline_days = 1
    assert occasion.deadline == (start - timedelta(days=2)).date()

    period.deadline_date = date(2017, 2, 23)
    assert occasion.deadline == date(2017, 2, 23)


def test_cancellation_deadline(session, collections, prebooking_period, owner):
    period = prebooking_period

    start, end = period.execution_start,\
        period.execution_start + timedelta(hours=2)

    occasion = collections.occasions.add(
        start=start,
        end=end,
        timezone="Europe/Zurich",
        activity=collections.activities.add("Sport", username=owner.username),
        period=period
    )
    assert occasion.cancellation_deadline is None

    period.cancellation_days = 1
    assert occasion.cancellation_deadline == (start - timedelta(days=2)).date()

    period.cancellation_date = date(2017, 2, 23)
    assert occasion.cancellation_deadline == date(2017, 2, 23)


def test_prebooking_phases():
    period = Period()

    period.prebooking_start = date(2017, 5, 1)
    period.prebooking_end = date(2017, 5, 2)

    with freeze_time('2017-04-30 23:59:59'):
        assert period.is_prebooking_in_future

    with freeze_time('2017-05-01 00:00:00'):
        assert not period.is_prebooking_in_future

        period.active = False
        assert not period.is_currently_prebooking

        period.active = True
        assert period.is_currently_prebooking

    with freeze_time('2017-05-05 00:00:00'):
        assert not period.is_prebooking_in_future
        assert not period.is_currently_prebooking
        assert period.is_prebooking_in_past

    with freeze_time('2017-05-02 23:59:59'):
        assert period.is_currently_prebooking

        period.confirmed = True
        assert not period.is_currently_prebooking
        assert period.is_prebooking_in_past


def test_publication_request(session, owner):

    activities = ActivityCollection(session)
    requests = PublicationRequestCollection(session)
    periods = PeriodCollection(session)

    period = periods.add(
        title="Autumn 2016",
        prebooking=(datetime(2016, 9, 1), datetime(2016, 9, 30)),
        execution=(datetime(2016, 10, 1), datetime(2016, 10, 31)),
        active=True
    )

    activity = activities.add(
        title="Sport",
        username=owner.username,
    )

    request = requests.add(activity=activity, period=period)
    session.flush()

    request = requests.query().first()
    assert request.activity.title == "Sport"
    assert request.period.title == "Autumn 2016"


def test_archive_period(session, owner):

    activities = ActivityCollection(session)
    occasions = OccasionCollection(session)
    periods = PeriodCollection(session)

    current_period = periods.add(
        title="Autumn 2017",
        prebooking=(datetime(2017, 9, 1), datetime(2017, 9, 30)),
        execution=(datetime(2017, 10, 1), datetime(2017, 10, 31)),
        active=True
    )

    future_period = periods.add(
        title="Winter 2017",
        prebooking=(datetime(2017, 11, 1), datetime(2017, 11, 30)),
        execution=(datetime(2017, 12, 1), datetime(2017, 12, 31)),
        active=False
    )

    sport = activities.add("Sport", username=owner.username)
    games = activities.add("Games", username=owner.username)
    empty = activities.add("Empty", username=owner.username)

    sport.propose().accept()
    games.propose().accept()
    empty.propose().accept()

    occasions.add(
        start=datetime(2017, 10, 4, 13),
        end=datetime(2017, 10, 4, 14),
        timezone="Europe/Zurich",
        meeting_point="Lucerne",
        age=(6, 9),
        spots=(2, 10),
        note="Bring game-face",
        activity=sport,
        period=current_period
    )
    occasions.add(
        start=datetime(2017, 12, 4, 13),
        end=datetime(2017, 12, 4, 14),
        timezone="Europe/Zurich",
        meeting_point="Lucerne",
        age=(6, 9),
        spots=(2, 10),
        note="Bring game-face",
        activity=sport,
        period=future_period
    )
    occasions.add(
        start=datetime(2017, 12, 4, 13),
        end=datetime(2017, 12, 4, 14),
        timezone="Europe/Zurich",
        meeting_point="Lucerne",
        age=(6, 9),
        spots=(2, 10),
        note="Bring game-face",
        activity=games,
        period=current_period
    )

    current_period.confirmed = True
    current_period.finalized = True

    current_period.archive()

    assert current_period.archived == True
    assert sport.state == 'accepted'
    assert games.state == 'archived'
    assert empty.state == 'archived'


def test_activity_filter_toggle():
    f = ActivityFilter(tags=['Foo'])

    assert not f.toggled(tag='Foo').tags
    assert f.toggled(tag='Bar').tags == {'Foo', 'Bar'}


def test_activity_period_filter(scenario):
    # an activity fully booked in the previous period...
    scenario.add_period(title="Prev", active=False, confirmed=True)
    scenario.add_activity(title="Running", state='accepted')
    scenario.add_occasion(spots=(0, 1))

    scenario.commit()
    scenario.refresh()

    scenario.add_attendee()
    scenario.add_booking(state='accepted')

    # ...and unbooked in the current period
    scenario.add_period(title="Next", active=True, confirmed=True)
    scenario.add_occasion(spots=(0, 1))

    scenario.commit()
    scenario.refresh()

    assert scenario.occasions[0].attendee_count == 1
    assert scenario.occasions[1].attendee_count == 0

    # free spots should now depend on the selected period
    a = scenario.c.activities
    a.query().count() == 2

    assert a.for_filter(period_id=scenario.periods[0].id).query().count() == 1
    assert a.for_filter(period_id=scenario.periods[1].id).query().count() == 1

    assert a.for_filter(available='none').query().count() == 1
    assert a.for_filter(available='few').query().count() == 1

    assert a.for_filter(
        period_id=scenario.periods[0].id,
        available='many'
    ).query().count() == 0

    assert a.for_filter(
        period_id=scenario.periods[1].id,
        available='none'
    ).query().count() == 0


def test_activity_cost_filter(scenario):
    scenario.add_period()

    scenario.add_activity()
    scenario.add_occasion(cost=0)

    scenario.add_activity()
    scenario.add_occasion(cost=50)

    scenario.add_activity()
    scenario.add_occasion(cost=100)

    a = scenario.c.activities

    assert a.for_filter(price_range=(0, 0)).query().count() == 1
    assert a.for_filter(price_range=(0, 50)).query().count() == 2
    assert a.for_filter(price_range=(50, 100)).query().count() == 2
    assert a.for_filter(price_range=(0, 100)).query().count() == 3
    assert a.for_filter(price_range=(100, 1000)).query().count() == 1
    assert a.for_filter(price_range=(101, 1000)).query().count() == 0

    scenario.latest_period.all_inclusive = False
    scenario.latest_period.booking_cost = 1

    assert a.for_filter(price_range=(0, 0)).query().count() == 0
    assert a.for_filter(price_range=(1, 1)).query().count() == 1
    assert a.for_filter(price_range=(1, 51)).query().count() == 2
    assert a.for_filter(price_range=(51, 101)).query().count() == 2
    assert a.for_filter(price_range=(1, 101)).query().count() == 3
    assert a.for_filter(price_range=(101, 1000)).query().count() == 1
    assert a.for_filter(price_range=(102, 1000)).query().count() == 0

    scenario.latest_period.all_inclusive = True

    assert a.for_filter(price_range=(0, 0)).query().count() == 1
    assert a.for_filter(price_range=(0, 50)).query().count() == 2
    assert a.for_filter(price_range=(50, 100)).query().count() == 2
    assert a.for_filter(price_range=(0, 100)).query().count() == 3
    assert a.for_filter(price_range=(100, 1000)).query().count() == 1
    assert a.for_filter(price_range=(101, 1000)).query().count() == 0


def test_timeline_filter(scenario):
    with freeze_time('2018-02-01'):
        scenario.add_period(active=False)
        scenario.add_activity(title="Winter")
        scenario.add_occasion()

    with freeze_time('2018-05-01'):
        scenario.add_period(active=False)
        scenario.add_activity(title="Spring")
        scenario.add_occasion()

    with freeze_time('2018-07-01'):
        scenario.add_period(active=False)
        scenario.add_activity(title="Summer")
        scenario.add_occasion()

    with freeze_time('2018-09-01'):
        scenario.add_period(active=False)
        scenario.add_activity(title="Autumn")

    a = scenario.c.activities

    with freeze_time('2018-01-01 23:30'):
        assert a.for_filter(timeline='past').query().count() == 0
        assert a.for_filter(timeline='now').query().count() == 0
        assert a.for_filter(timeline='future').query().count() == 3
        assert a.for_filter(timeline='undated').query().count() == 1

    with freeze_time('2018-02-10 23:30'):
        assert a.for_filter(timeline='past').query().count() == 0
        assert a.for_filter(timeline='now').query().count() == 1
        assert a.for_filter(timeline='future').query().count() == 2
        assert a.for_filter(timeline='undated').query().count() == 1

    with freeze_time('2018-04-01 23:30'):
        assert a.for_filter(timeline='past').query().count() == 1
        assert a.for_filter(timeline='future').query().count() == 2
        assert a.for_filter(timeline='undated').query().count() == 1

    with freeze_time('2018-05-10 22:30'):
        assert a.for_filter(timeline='past').query().count() == 1
        assert a.for_filter(timeline='now').query().count() == 1
        assert a.for_filter(timeline='future').query().count() == 1
        assert a.for_filter(timeline='undated').query().count() == 1

    with freeze_time('2018-06-01 22:30'):
        assert a.for_filter(timeline='past').query().count() == 2
        assert a.for_filter(timeline='future').query().count() == 1
        assert a.for_filter(timeline='undated').query().count() == 1

    with freeze_time('2018-07-10 22:30'):
        assert a.for_filter(timeline='past').query().count() == 2
        assert a.for_filter(timeline='now').query().count() == 1
        assert a.for_filter(timeline='future').query().count() == 0
        assert a.for_filter(timeline='undated').query().count() == 1

    with freeze_time('2018-08-01 22:30'):
        assert a.for_filter(timeline='past').query().count() == 3
        assert a.for_filter(timeline='future').query().count() == 0
        assert a.for_filter(timeline='undated').query().count() == 1


def test_no_occasion_in_period_filter(scenario):
    # an activity with an occasion in one period
    scenario.add_period(title="Previous", active=False, confirmed=True)
    scenario.add_activity(title="Running", state='accepted')
    scenario.add_occasion()

    # ...and no occasion in another
    scenario.add_period(title="Current", active=True, confirmed=True)
    scenario.add_activity(title="Swimming", state='accepted')

    scenario.commit()
    scenario.refresh()

    a = scenario.c.activities.for_filter(timeline='undated')

    # there is one undated activity
    assert a.query().count() == 1

    # 1 activity has no occasion in the previous period
    assert a.for_filter(period_id=scenario.periods[0].id).query().count() == 1

    # 2 activites have no occasion in the current period
    assert a.for_filter(period_id=scenario.periods[1].id).query().count() == 2



from onegov.activity.models import Occasion, OccasionDate
from onegov.core.collection import GenericCollection
from psycopg2.extras import NumericRange
from sedate import standardize_date


class OccasionCollection(GenericCollection):

    @property
    def model_class(self):
        return Occasion

    @staticmethod
    def to_half_open_interval(lower, upper):
        """ Postgres coerces ranges internally to be half-open in an effort
        to canonize these ranges. This function does the same by taking
        a closed interval and turning it into a half-open interval of
        the NumericRange type.

        """
        return NumericRange(lower, upper + 1, bounds='[)')

    def add(self, activity, period, start, end, timezone,
            meeting_point=None, age=None, spots=None, note=None, cost=0,
            exclude_from_overlap_check=False):

        start = standardize_date(start, timezone)
        end = standardize_date(end, timezone)

        occasion = super().add(
            meeting_point=meeting_point,
            age=age and self.to_half_open_interval(*age),
            spots=spots and self.to_half_open_interval(*spots),
            note=note,
            activity_id=activity.id,
            period_id=period.id,
            cost=cost,
            exclude_from_overlap_check=exclude_from_overlap_check
        )

        self.add_date(occasion, start, end, timezone)

        return occasion

    def add_date(self, occasion, start, end, timezone):

        start = standardize_date(start, timezone)
        end = standardize_date(end, timezone)

        date = OccasionDate(
            start=start,
            end=end,
            timezone=timezone
        )

        occasion.dates.append(date)
        self.session.flush()

        return date

    def find_date(self, occasion, start, end, timezone):
        start = standardize_date(start, timezone)
        end = standardize_date(end, timezone)

        q = self.session.query(OccasionDate)
        q = q.filter(OccasionDate.start == start)
        q = q.filter(OccasionDate.end == end)
        q = q.filter(OccasionDate.timezone == timezone)

        return q.first()

    def remove_date(self, date):
        q = self.session.query(OccasionDate)
        q = q.filter(OccasionDate.id == date.id)

        date = q.first()

        if date:
            self.session.delete(date)

    def clear_dates(self, occasion):
        q = self.session.query(OccasionDate)
        q = q.filter(OccasionDate.occasion_id == occasion.id)

        for date in q:
            self.session.delete(date)

        occasion.dates = []
        self.session.flush()



from onegov.activity.matching.core import deferred_acceptance
from onegov.activity.matching.core import deferred_acceptance_from_database
from onegov.activity.matching.interfaces import MatchableBooking
from onegov.activity.matching.interfaces import MatchableOccasion
from onegov.activity.matching.score import PreferAdminChildren
from onegov.activity.matching.score import PreferGroups
from onegov.activity.matching.score import PreferInAgeBracket
from onegov.activity.matching.score import PreferMotivated
from onegov.activity.matching.score import PreferOrganiserChildren
from onegov.activity.matching.score import Scoring

__all__ = [
    'deferred_acceptance',
    'deferred_acceptance_from_database',
    'MatchableBooking',
    'MatchableOccasion',
    'Scoring',
    'PreferGroups',
    'PreferMotivated',
    'PreferInAgeBracket',
    'PreferOrganiserChildren',
    'PreferAdminChildren',
]



import re
import secrets
import string

from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.core.utils import chunks, hash_dictionary
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Text
from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import backref, relationship, validates


KNOWN_SCHEMAS = {}

INVALID_REFERENCE_CHARS_EX = re.compile(r'[^Q0-9A-F]+')
REFERENCE_EX = re.compile(r'Q{1}[A-F0-9]{10}')


class InvoiceReference(Base, TimestampMixin):
    """ A reference pointing to an invoice. References are keys which are used
    outside the application. Usually a code used on an invoice to enter through
    online-banking.

    Each invoice may have multiple references pointing to it. Each refernce
    is however unique.

    There are multiple schemas for references. Each schema generates its own
    set of references using a Python class and some optional settings given
    by the user (say a specific bank's account number).

    Each schema may only reference an invoice once.

    Though each schema has its own set of references, the references-space
    is shared between all schemas. In other words, reference 'foo' of schema
    A would conflict with reference 'foo' of schema B.

    This is because we do not know which schema was used when we encounter
    a reference.

    In reality this should not be a problem as reference schemes provided
    by banks usually cover a very large space, so multiple schemas are expected
    to just generate random numbers until one is found that has not been used
    yet (which should almost always happen on the first try).

    """

    __tablename__ = 'invoice_references'

    #: the unique reference
    reference = Column(Text, primary_key=True)

    #: the referenced invoice
    invoice_id = Column(UUID, ForeignKey('invoices.id'), nullable=False)
    invoice = relationship(
        'Invoice', backref=backref(
            "references", cascade="all, delete-orphan"))

    #: the schema used to generate the invoice
    schema = Column(Text, nullable=False)

    #: groups schema name and its config to identify records created by a
    #: given schema and config
    bucket = Column(Text, nullable=False)

    __table_args__ = (
        UniqueConstraint(
            'bucket', 'invoice_id', name='unique_bucket_invoice_id'),
    )

    @validates
    def validate_schema(self, field, value):

        if value not in KNOWN_SCHEMAS:
            raise ValueError(f'{value} is an unknown schema')

        return value

    @property
    def readable(self):
        """ Returns the human formatted variant of the reference. """

        return KNOWN_SCHEMAS[self.schema]().format(self.reference)


class Schema(object):
    """ Defines the methods that need to be implemented by schemas. Schemas
    should generate numbers and be able to format them.

    Schemas should never be deleted as we want to be able to display
    past schemas even if a certain schema is no longer in use.

    If a new schema comes along that replaces an old one in an incompatible
    way, the new schema should get a new name and should be added alongside
    the old one.

    """

    def __init_subclass__(cls, name, **kwargs):
        super().__init_subclass__(**kwargs)

        cls.name = name
        KNOWN_SCHEMAS[cls.name] = cls

    def __init__(self, **config):
        """ Each schema may have a custom config. This is *only used for
        creation of references*. For other uses like formatting the config
        is not passed in.

        """
        for k, v in config.items():
            setattr(self, k, v)

        self.config = config

    @property
    def bucket(self):
        """ Generates a unique identifer for the current schema and config. """

        return self.render_bucket(self.name, self.config)

    @classmethod
    def render_bucket(cls, schema_name, schema_config=None):
        if schema_config:
            return f'{schema_name}-{hash_dictionary(schema_config)}'

        return schema_name

    def link(self, session, invoice, optimistic=False, flush=True):
        """ Creates a new :class:`InvoiceReference` for the given invoice.

        The returned invoice should have a unique reference, so the chance
        of ending up with a conflict error later down the line are slim.

        If the schema already has a linke to the invoice, we skip the
        creation.

        By default we check our constraints before we write to the database.
        To be faster in performance critical situation we can however also
        chose to be 'optimistic' and forego those checks. Due to the random
        nature of schema references this should usually work.

        The constraints are set on the database, so they will be enforced
        either way.

        Additionally we can forego the session.flush if we want to.

        """

        assert invoice.id, "the invoice id must be konwn"

        q = optimistic or session.query(InvoiceReference)

        # check if we are already linked
        if not optimistic:
            if q.filter_by(bucket=self.bucket, invoice_id=invoice.id).first():
                return

        # find an unused reference
        for i in range(0, 10):
            candidate = self.new()

            if not optimistic:
                if q.filter_by(reference=candidate).first():
                    continue

            break
        else:
            raise RuntimeError("No unique reference after 10 tries")

        reference = InvoiceReference(
            invoice_id=invoice.id,
            reference=candidate,
            schema=self.name,
            bucket=self.bucket,
        )

        session.add(reference)

        if flush:
            session.flush()

        return reference

    def new(self):
        """ Returns a new reference in the most compact way possible. """
        raise NotImplementedError()

    def format(self, reference):
        """ Turns the reference into something human-readable. """
        raise NotImplementedError()


class FeriennetSchema(Schema, name='feriennet-v1'):
    """ The default schema for customers without specific bank integrations.

    The generated reference is entered as a note when conducting the
    online-banking transaction.

    """

    def new(self):
        return f'q{secrets.token_hex(5)}'

    def format(self, reference):
        reference = reference.upper()

        return '-'.join((
            reference[:1],
            reference[1:6],
            reference[6:]
        ))

    def extract(self, text):
        """ Takes a bunch of text and tries to extract the feriennet-v1
        reference from it.

        """
        if text is None:
            return None

        text = text.replace('\n', '').strip()

        if not text:
            return None

        # ENTER A WORLD WITHOUT LOWERCASE
        text = text.upper()

        # replace all O-s (as in OMG) with 0.
        text = text.replace('O', '0')

        # normalize the text by removing all invalid characters.
        text = INVALID_REFERENCE_CHARS_EX.sub('', text)

        # try to fetch the reference
        match = REFERENCE_EX.search(text)

        if not match:
            return None

        return match.group().lower()


class ESRSchema(Schema, name='esr-v1'):
    """ The default schema for ESR by Postfinance. In it's default form it is
    random and requires no configuration.

    A ESR reference has 27 characters from 0-9. The first 26 characters can
    be chosen freely and the last character is the checksum.

    Some banks require that references have certain prefixes/suffixes, but
    Postfinance who defined the ESR standard does not.

    """

    def new(self):
        number = ''.join(secrets.choice(string.digits) for _ in range(0, 26))
        return number + self.checksum(number)

    def checksum(self, number):
        """ Generates the modulo 10 checksum as required by Postfinance. """

        table = (0, 9, 4, 6, 8, 2, 7, 1, 3, 5)
        carry = 0

        for n in number:
            carry = table[(carry + int(n)) % 10]

        return str((10 - carry) % 10)

    def format(self, reference):
        """ Takes an ESR reference and formats it in a human-readable way.

        This is mandated as follows by Postfinance:

        > Die Referenznummer ist rechtsbündig, in 5er-Blocks und einem
        > allfälligen Restblock zu platzieren. Vorlaufende Nullen können
        > unterdrückt werden.

        """

        reference = reference.lstrip('0')
        reference = reference.replace(' ', '')

        blocks = []

        for values in chunks(reversed(reference), n=5, fillvalue=''):
            blocks.append(''.join(reversed(values)))

        return ' '.join(reversed(blocks))


class RaiffeisenSchema(ESRSchema, name='raiffeisen-v1'):
    """ Customised ESR for Raiffeisen. """

    def new(self):
        ident = self.esr_identification_number.replace('-', '').strip()
        assert 3 <= len(ident) <= 7

        rest = 26 - len(ident)
        random = ''.join(secrets.choice(string.digits) for _ in range(0, rest))
        number = f'{self.esr_identification_number}{random}'

        return number + self.checksum(number)



from onegov.activity.models import Booking, Period
from onegov.core.collection import GenericCollection
from onegov.activity.matching.utils import unblockable, booking_order
from onegov.activity.errors import BookingLimitReached
from sqlalchemy.orm import joinedload


class BookingCollection(GenericCollection):

    def __init__(self, session, period_id=None, username=None):
        super().__init__(session)
        self.period_id = period_id
        self.username = username

    def query(self):
        query = super().query()

        if self.username is not None:
            query = query.filter(Booking.username == self.username)

        if self.period_id is not None:
            query = query.filter(Booking.period_id == self.period_id)

        return query

    def for_period(self, period):
        return self.__class__(self.session, period.id, self.username)

    def for_username(self, username):
        return self.__class__(self.session, self.period_id, username)

    @property
    def model_class(self):
        return Booking

    def count(self, usernames='*', periods='*', states='*'):
        """ Returns the number of bookings, optionally filtered by usernames,
        periods and states.

        All parameters may either be iterables or subqueries.

        """

        query = self.query().with_entities(Booking.id)

        if states != '*':
            query = query.filter(Booking.state.in_(states))

        if periods != '*':
            query = query.filter(Booking.period_id.in_(periods))

        if usernames != '*':
            query = query.filter(Booking.username.in_(usernames))

        return query.count()

    def booking_count(self, username, states='*'):
        """ Returns the number of bookings in the active period. """

        periods = self.session.query(Period)
        periods = periods.with_entities(Period.id)
        periods = periods.filter(Period.active == True)

        return self.count(
            usernames=(username, ),
            periods=periods.subquery(),
            states=states
        )

    def by_user(self, user):
        return self.query().filter(Booking.username == user.username)

    def by_username(self, username):
        return self.query().filter(Booking.username == username)

    def by_occasion(self, occasion):
        return self.query().filter(Booking.occasion_id == occasion.id)

    def add(self, user, attendee, occasion, priority=None, group_code=None):

        return super().add(
            username=user.username,
            attendee_id=attendee.id,
            occasion_id=occasion.id,
            priority=priority,
            group_code=group_code,
            period_id=occasion.period_id
        )

    def accept_booking(self, booking):
        """ Accepts the given booking, setting all other bookings which
        conflict with it to 'blocked'.

        Can only be done if the period has been confirmed already, the
        occasion is not yet full and the given booking doesn't conflict
        with another accepted booking.

        """

        if not booking.period.confirmed:
            raise RuntimeError("The period has not yet been confirmed")

        if booking.occasion.full:
            raise RuntimeError("The occasion is already full")

        if booking.state not in ('open', 'denied'):
            raise RuntimeError("Only open/denied bookings can be accepted")

        bookings = tuple(
            self.session.query(Booking)
                .options(joinedload(Booking.occasion))
                .filter(Booking.attendee_id == booking.attendee_id)
                .filter(Booking.period_id == booking.period_id)
                .filter(Booking.id != booking.id)
        )

        limit = booking.attendee.limit or booking.period.booking_limit

        if limit and not booking.occasion.exempt_from_booking_limit:
            accepted = sum(
                1 for b in bookings if b.state == 'accepted'
                and not b.occasion.exempt_from_booking_limit
            )

            if accepted >= limit:
                raise BookingLimitReached()

            # accepting one more booking will reach the limit
            block_rest = (accepted + 1) >= limit
        else:
            block_rest = False

        # block the overlapping bookings
        for b in bookings:
            if b.overlaps(booking):
                if b.state == 'cancelled':
                    continue

                if b.state == 'accepted':
                    raise RuntimeError("Conflict with booking {}".format(b.id))

                b.state = 'blocked'

        # if we reached the limit, block *all* bookings
        if block_rest:
            for b in bookings:
                if b.state in ('cancelled', 'accepted'):
                    continue

                if b.occasion.exempt_from_booking_limit:
                    continue

                b.state = 'blocked'

        booking.state = 'accepted'
        booking.cost = booking.provisional_booking_cost()

    def cancel_booking(self, booking,
                       score_function=booking_order, cascade=True):
        """ Cancels the given booking.

        If ``cascade`` is set to False, this amounts to a simple state change
        and you can stop reading now.

        If ``cascade`` is set to true, all denied bookings which have a chance
        of becoming accepted as a result (because the occasion frees up) are
        accepted according to their matching score (already accepted bookings
        are not touched).

        All bookings which get unblocked for the current user are tried
        to be accepted as well, also with the highest score first. Contrary
        to the other influenced bookings, these bookings are not limited to
        the occasion of the cancelled booking.

        This won't necesserily create a new stable matching, but it will
        keep the operability as high as possible. It's not ideal from a
        usability perspective as a single cancel may have a bigger impact
        than the user intended, but it beats having occasions fail with
        too few attendees when there are bookings waiting.

        Can only be done if the period has been confirmed already and the
        booking is an accepted bookings. Open, cancelled, blocked and
        denied bookings can simply be deleted.

        """

        if not booking.period.confirmed:
            raise RuntimeError("The period has not yet been confirmed")

        # if the booking wasn't accepted or if we don't cascade, this is quick
        if not cascade or booking.state != 'accepted':
            booking.state = 'cancelled'
            booking.group_code = None
            return

        bookings = tuple(
            self.session.query(Booking)
                .options(joinedload(Booking.occasion))
                .filter(Booking.attendee_id == booking.attendee_id)
                .filter(Booking.period_id == booking.period_id)
                .filter(Booking.id != booking.id))

        booking.state = 'cancelled'
        booking.group_code = None

        # mark the no-longer blocked bookings as denied
        accepted = {b for b in bookings if b.state == 'accepted'}
        blocked = {b for b in bookings if b.state == 'blocked'}
        unblocked = set()

        if booking.period.all_inclusive:
            limit = booking.period.booking_limit
        else:
            limit = booking.attendee.limit

        unblockable_bookings = unblockable(accepted, blocked, score_function)

        for cnt, booking in enumerate(unblockable_bookings, start=1):

            if limit and limit < (cnt + len(accepted)):
                break

            booking.state = 'denied'
            unblocked.add(booking)

        self.session.flush()

        # try to accept the denied bookings in their respective occasions
        for b in unblocked:

            # the denied state changes during the loop execution
            if b.state == 'denied' and not b.occasion.full:
                self.accept_booking(b)
                self.session.flush()

        # try to accept the open/denied bookings in the current occasion
        spots = booking.occasion.available_spots
        denied_bookings = sorted(
            (
                b for b in booking.occasion.bookings
                if b.state in ('open', 'denied')
            ),
            key=score_function)

        for b in denied_bookings:
            if spots:
                try:
                    self.accept_booking(b)
                    self.session.flush()
                    spots -= 1
                except BookingLimitReached:
                    pass



import re

from cached_property import cached_property
from collections import defaultdict
from datetime import date
from decimal import Decimal
from lxml import etree
from onegov.activity.collections import InvoiceCollection
from onegov.activity.models import Invoice
from onegov.activity.models import InvoiceItem
from onegov.activity.models import InvoiceReference
from onegov.activity.models.invoice_reference import FeriennetSchema
from onegov.user import User
from sqlalchemy import func


DOCUMENT_NS_EX = re.compile(r'.*<Document [^>]+>(.*)')


def normalize_xml(xml):
    # let's not bother with namespaces at all
    return DOCUMENT_NS_EX.sub(r'<Document>\1', xml)


class Transaction(object):

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.username = None
        self.confidence = 0
        self.duplicate = False
        self.paid = False

    def __repr__(self):
        return repr(self.__dict__)

    @cached_property
    def references(self):
        return set(self.extract_references())

    def extract_references(self):
        if self.reference:
            yield self.reference

        # currently only one schema supports text extraction, the others are
        # stored in the designated reference field
        schema = FeriennetSchema()

        ref = schema.extract(self.booking_text)
        if ref:
            yield ref

        ref = schema.extract(self.note)
        if ref:
            yield ref

    @property
    def order(self):
        state = self.state

        if self.valuta_date:
            date = self.valuta_date and (
                self.valuta_date.year * 10000
                + self.valuta_date.month * 100
                + self.valuta_date.day
            ) * - 1
        else:
            date = -1

        if state == 'success':
            return 0, date, self.username

        if state == 'paid':
            return 1, date, self.username

        if state == 'duplicate':
            return 2, date, self.username

        if state == 'warning':
            return 3, date, self.username

        return 4, date

    @property
    def state(self):
        if self.paid:
            return 'paid'

        if self.duplicate:
            return 'duplicate'

        if self.confidence == 1:
            return 'success'

        if self.confidence == 0.5:
            return 'warning'

        return 'unknown'


def transaction_entries(root):
    """ Yields the transaction entries from the given Camt.053 or Camt.054
    xml. This works because for our purposes the entries of those two formats
    are identical.

    """

    # camt.053
    for entry in root.xpath('/Document/BkToCstmrStmt/Stmt/Ntry'):
        yield entry

    # camt.054
    for entry in root.xpath('/Document/BkToCstmrDbtCdtNtfctn/Ntfctn/Ntry'):
        yield entry


def extract_transactions(xml):
    root = etree.fromstring(normalize_xml(xml).encode('utf-8'))

    def first(element, xpath):
        elements = element.xpath(xpath)
        return elements[0] if elements else None

    def joined(element, xpath):
        return '\n'.join(element.xpath(xpath))

    def as_decimal(text):
        if text:
            return Decimal(text)

    def as_date(text):
        if text:
            return date(*[int(p) for p in text.split('-')])

    for entry in transaction_entries(root):
        booking_date = as_date(first(entry, 'BookgDt/Dt/text()'))
        valuta_date = as_date(first(entry, 'ValDt/Dt/text()'))
        booking_text = first(entry, 'AddtlNtryInf/text()')

        for d in entry.xpath('NtryDtls/TxDtls'):
            yield Transaction(
                booking_date=booking_date,
                valuta_date=valuta_date,
                booking_text=booking_text,
                tid=first(d, 'Refs/AcctSvcrRef/text()'),
                amount=as_decimal(first(d, 'Amt/text()')),
                currency=first(d, 'Amt/@Ccy'),
                reference=first(d, 'RmtInf/Strd/CdtrRefInf/Ref/text()'),
                note=joined(d, 'RmtInf/Ustrd/text()'),
                credit=first(d, 'CdtDbtInd/text()') == 'CRDT',
                debitor=first(d, 'RltdPties/Dbtr/Nm/text()'),
                debitor_account=first(d, 'RltdPties/DbtrAcct/Id/IBAN/text()'),
            )


def match_iso_20022_to_usernames(xml, session, period_id, currency='CHF'):
    """ Takes an ISO20022 camt.053 file and matches it with the invoice
    items in the database.

    Raises an error if the given xml cannot be processed.

    :return: A list of transactions found in the xml file, together with
    the matching username and a confidence attribute indicating how
    certain the match is (1.0 indicating a sure match, 0.5 a possible match
    and 0.0 a non-match).

    """

    def items(period_id=None):
        invoices = InvoiceCollection(session, period_id=period_id)
        return invoices.query_items().outerjoin(Invoice).outerjoin(User)

    # Get all known transaction ids to check what was already paid
    q = items()
    q = q.with_entities(InvoiceItem.tid, User.username)
    q = q.group_by(InvoiceItem.tid, User.username)
    q = q.filter(
        InvoiceItem.paid == True,
        InvoiceItem.source == 'xml'
    )

    paid_transaction_ids = {i.tid: i.username for i in q}

    # Get a list of reference/username pairs as fallback
    q = session.query(InvoiceReference).outerjoin(Invoice).outerjoin(User)
    q = q.with_entities(InvoiceReference.reference, User.username)

    username_by_ref = dict(q)

    # Get the items matching the given period
    q = items(period_id=period_id).outerjoin(InvoiceReference)
    q = q.with_entities(
        User.username,
        func.sum(InvoiceItem.amount).label('amount'),
        InvoiceReference.reference,
    )
    q = q.group_by(User.username, InvoiceReference.reference)
    q = q.filter(InvoiceItem.paid == False)
    q = q.order_by(User.username)

    # Hash the invoices by reference (duplicates possible)
    by_ref = defaultdict(list)

    # Hash the invoices by amount (duplicates probable)
    by_amount = defaultdict(list)

    last_username = None

    for record in q:
        by_ref[record.reference].append(record)

        if last_username != record.username:
            by_amount[record.amount].append(record)
            last_username = record.username

    # go through the transactions, comparing amount and code for a match
    transactions = tuple(extract_transactions(xml))

    # mark duplicate transactions
    seen = {}

    for t in transactions:
        for ref in t.references:
            if ref in seen:
                t.duplicate = seen[ref].duplicate = True

            seen[ref] = t

    for t in transactions:

        # credit transactions are completely irrelevant for us
        if not t.credit:
            continue

        if t.currency != currency:
            yield t
            continue

        if t.tid in paid_transaction_ids:
            t.paid = True
            t.username = paid_transaction_ids[t.tid]
            t.confidence = 1
            yield t
            continue

        if t.credit and t.currency == currency:
            amnt_usernames = {i.username for i in by_amount[t.amount]}
            ref_usernames = {
                i.username for ref in t.references for i in by_ref[ref]}

            combined = amnt_usernames & ref_usernames

            if len(combined) == 1:
                t.username = next(u for u in combined)
                t.confidence = 1.0
            elif len(ref_usernames) == 1:
                t.username = next(u for u in ref_usernames)
                t.confidence = 0.5
            elif len(amnt_usernames) == 1:
                t.username = next(u for u in amnt_usernames)
                t.confidence = 0.5

            if not t.confidence:
                for ref in t.references:
                    if ref in username_by_ref:
                        t.username = username_by_ref[ref]
                        t.confidence = 0.5

        yield t



from cached_property import cached_property
from decimal import Decimal
from onegov.activity.models import Invoice, InvoiceItem
from onegov.activity.models.invoice_reference import KNOWN_SCHEMAS
from onegov.core.collection import GenericCollection
from sqlalchemy import func, and_, not_
from onegov.user import User
from uuid import uuid4


class InvoiceCollection(GenericCollection):

    def __init__(self, session, period_id=None, user_id=None,
                 schema='feriennet-v1', schema_config=None):
        super().__init__(session)
        self.user_id = user_id
        self.period_id = period_id

        if schema not in KNOWN_SCHEMAS:
            raise RuntimeError("Unknown schema: {schema}")

        self.schema_name = schema
        self.schema_config = (schema_config or {})

    @cached_property
    def schema(self):
        return KNOWN_SCHEMAS[self.schema_name](**self.schema_config)

    def query(self):
        q = super().query()

        if self.user_id:
            q = q.filter_by(user_id=self.user_id)

        if self.period_id is not None:
            q = q.filter_by(period_id=self.period_id)

        return q

    def query_items(self):
        return self.session.query(InvoiceItem)\
            .filter(InvoiceItem.invoice_id.in_(
                self.query().with_entities(Invoice.id).subquery()
            ))

    def for_user_id(self, user_id):
        return self.__class__(self.session, self.period_id, user_id,
                              self.schema_name, self.schema_config)

    def for_period_id(self, period_id):
        return self.__class__(self.session, period_id, self.user_id,
                              self.schema_name, self.schema_config)

    def for_schema(self, schema, schema_config=None):
        return self.__class__(self.session, self.period_id, self.user_id,
                              schema, schema_config)

    @cached_property
    def invoice(self):
        # XXX used for compatibility with legacy implementation in Feriennet
        return self.period_id and self.period_id.hex or None

    @cached_property
    def username(self):
        # XXX used for compatibility with legacy implementation in Feriennet
        if self.user_id:
            user = self.session.query(User)\
                .with_entities(User.username)\
                .filter_by(id=self.user_id)\
                .first()

            return user and user.username or None

    @property
    def model_class(self):
        return Invoice

    def _invoice_ids(self):
        return self.query().with_entities(Invoice.id).subquery()

    def _sum(self, condition):
        q = self.session.query(func.sum(InvoiceItem.amount).label('amount'))
        q = q.filter(condition)

        return Decimal(q.scalar() or 0.0)

    @property
    def total_amount(self):
        return self._sum(InvoiceItem.invoice_id.in_(self._invoice_ids()))

    @property
    def outstanding_amount(self):
        return self._sum(and_(
            InvoiceItem.invoice_id.in_(self._invoice_ids()),
            InvoiceItem.paid == False
        ))

    @property
    def paid_amount(self):
        return self._sum(and_(
            InvoiceItem.invoice_id.in_(self._invoice_ids()),
            InvoiceItem.paid == True
        ))

    def unpaid_count(self, excluded_period_ids=None):
        q = self.query().with_entities(func.count(Invoice.id))

        if excluded_period_ids:
            q = q.filter(not_(Invoice.period_id.in_(excluded_period_ids)))

        q = q.filter(Invoice.paid == False)

        return q.scalar() or 0

    def sync(self):
        for invoice in self.query():
            invoice.sync()

    def add(self, period_id=None, user_id=None, flush=True, optimistic=False):
        invoice = Invoice(
            id=uuid4(),
            period_id=period_id or self.period_id,
            user_id=user_id or self.user_id)

        self.session.add(invoice)
        self.schema.link(
            self.session, invoice, flush=flush, optimistic=optimistic)

        if flush:
            self.session.flush()

        return invoice



import secrets
import sedate

from copy import copy
from enum import IntEnum
from onegov.activity.models import Activity, Occasion, OccasionDate, Period
from onegov.activity.utils import date_range_decode
from onegov.activity.utils import date_range_encode
from onegov.activity.utils import merge_ranges
from onegov.activity.utils import num_range_decode
from onegov.activity.utils import num_range_encode
from onegov.activity.utils import overlaps
from onegov.core.collection import Pagination
from onegov.core.utils import increment_name
from onegov.core.utils import is_uuid
from onegov.core.utils import normalize_for_url
from onegov.core.utils import toggle
from sedate import utcnow
from sqlalchemy import and_, or_, not_
from sqlalchemy import column
from sqlalchemy import distinct
from sqlalchemy import exists
from sqlalchemy import func
from sqlalchemy import select
from sqlalchemy import text
from sqlalchemy.dialects.postgresql import array
from uuid import UUID


AVAILABILITY_VALUES = {'none', 'few', 'many'}


class ActivityFilter(object):

    # supported filters - should be named with a plural version that can
    # be turned into a singular with the removal of the last s
    # (i.e. slots => slot)
    __slots__ = (
        'age_ranges',
        'available',
        'price_ranges',
        'dateranges',
        'durations',
        'municipalities',
        'owners',
        'period_ids',
        'states',
        'tags',
        'timelines',
        'weekdays',
    )

    singular = {
        'municipalities': 'municipality'
    }

    def __init__(self, **keywords):

        for key in self.__slots__:
            if key in keywords:
                values = set(keywords[key]) if keywords[key] else set()

                if values and hasattr(self, f'adapt_{key}'):
                    try:
                        values = getattr(self, f'adapt_{key}')(values)
                    except ValueError:
                        values = set()

                setattr(self, key, values)
            else:
                setattr(self, key, set())

    @property
    def keywords(self):
        keywords = {}

        for key in self.__slots__:
            if getattr(self, key):
                keywords[key] = self.encode(key, getattr(self, key))

        return keywords

    def toggled(self, **keywords):
        # create a new filter with the toggled values
        toggled = copy(self)

        for key in self.__slots__:

            # support plural and singular
            if key.endswith('s'):
                singular = self.singular.get(key, key[:-1])
            else:
                singular = key

            if singular in keywords:
                value = keywords[singular]
            elif key in keywords:
                value = keywords[key]
            else:
                continue

            setattr(toggled, key, toggle(getattr(self, key), value))

        return toggled

    def adapt_available(self, values):
        return values & AVAILABILITY_VALUES

    def adapt_num_ranges(self, values):
        decoded = set(v for v in map(num_range_decode, values) if v)
        return decoded and set(merge_ranges(decoded)) or set()

    def adapt_age_ranges(self, values):
        return self.adapt_num_ranges(values)

    def adapt_price_ranges(self, values):
        return self.adapt_num_ranges(values)

    def adapt_dateranges(self, values):
        return set(v for v in map(date_range_decode, values) if v)

    def adapt_weekdays(self, values):
        return set(int(v) for v in values if v.isdigit())

    def adapt_period_ids(self, values):
        return set(UUID(v) for v in values if is_uuid(v))

    def adapt_durations(self, values):
        return set(int(v) for v in values)

    def encode(self, key, value):
        if isinstance(value, str):
            return value

        if key == 'dateranges':
            return [date_range_encode(v) for v in value]

        if key == 'age_ranges':
            return [num_range_encode(v) for v in value]

        if key == 'price_ranges':
            return [num_range_encode(v) for v in value]

        if isinstance(value, IntEnum):
            return str(int(value))

        if isinstance(value, int):
            return str(value)

        if isinstance(value, UUID):
            return value.hex

        if isinstance(value, (tuple, list, set)):
            return [self.encode(key, v) for v in value]

        raise NotImplementedError()

    def contains_num_range(self, value, ranges):
        for r in ranges:
            if overlaps(r, value):
                return True
        return False

    def contains_age_range(self, age_range):
        return self.contains_num_range(age_range, self.age_ranges)

    def contains_price_range(self, price_range):
        return self.contains_num_range(price_range, self.price_ranges)


class ActivityCollection(Pagination):

    def __init__(self, session, type='*', page=0, filter=None):
        self.session = session
        self.type = type
        self.page = page
        self.filter = filter or ActivityFilter()

    def __eq__(self, other):
        return self.type == other.type and self.page == other.page

    def subset(self):
        return self.query()

    @property
    def page_index(self):
        return self.page

    def page_by_index(self, index):
        return self.__class__(
            self.session,
            type=self.type,
            page=index,
            filter=self.filter
        )

    @property
    def model_class(self):
        return Activity.get_polymorphic_class(self.type, Activity)

    def query_base(self):
        """ Returns the query based used by :meth:`query`. Overriding this
        function is useful to apply a general filter to the query before
        any other filter is applied.

        For example, a policy can be enforced that only allows public
        activites.

        """
        return self.session.query(self.model_class)

    def query(self):
        query = self.query_base()
        model_class = self.model_class

        # activity based filters
        if self.type != '*':
            query = query.filter(model_class.type == self.type)

        if self.filter.tags:
            query = query.filter(
                model_class._tags.has_any(array(self.filter.tags)))

        if self.filter.states:
            query = query.filter(
                model_class.state.in_(self.filter.states))

        if self.filter.owners:
            query = query.filter(model_class.username.in_(
                self.filter.owners))

        if self.filter.municipalities:
            query = query.filter(
                model_class.municipality.in_(self.filter.municipalities))

        # occasion based filters
        o = self.session.query(Occasion)\
            .with_entities(Occasion.activity_id)\
            .join(OccasionDate)\
            .distinct()

        # if we are looking at activities without occasions, we do not have
        # to apply all the filters below which are occasion-based
        if 'undated' in self.filter.timelines:
            if self.filter.period_ids:
                return query.filter(
                    not_(self.session.query(Occasion.activity_id).filter(and_(
                        Occasion.activity_id == Activity.id,
                        Occasion.period_id.in_(self.filter.period_ids)
                    )).exists())
                )
            else:
                return query.filter(
                    not_(self.session.query(Occasion.activity_id).filter(
                        Occasion.activity_id == Activity.id,
                    ).exists())
                )

        now = utcnow()

        if self.filter.timelines:
            conditions = []

            if 'past' in self.filter.timelines:
                conditions.append(OccasionDate.end < now)

            if 'now' in self.filter.timelines:
                conditions.append(and_(
                    OccasionDate.start <= now, now <= OccasionDate.end
                ))

            if 'future' in self.filter.timelines:
                conditions.append(now < OccasionDate.start)

            # the 'undated' option can only be implemented further down
            # when we apply the occasion conditions to the activites query
            # since we'd be looking for activities without occasions
            if conditions:
                o = o.filter(or_(*conditions))

        if self.filter.period_ids:
            o = o.filter(Occasion.period_id.in_(self.filter.period_ids))

        if self.filter.durations:
            o = o.filter(Occasion.duration.in_(
                int(d) for d in self.filter.durations))

        if self.filter.age_ranges:
            o = o.filter(or_(
                *(
                    Occasion.age.overlaps(func.int4range(min_age, max_age + 1))
                    for min_age, max_age in self.filter.age_ranges
                )
            ))

        if self.filter.price_ranges:
            o = o.join(Period)
            o = o.filter(or_(
                *(
                    and_(
                        min_price <= (
                            func.coalesce(Occasion.cost, 0)
                            + func.coalesce(Period.occasion_extra_cost, 0)
                        ),
                        (
                            func.coalesce(Occasion.cost, 0)
                            + func.coalesce(Period.occasion_extra_cost, 0)
                        ) <= max_price
                    ) for min_price, max_price in self.filter.price_ranges
                )
            ))

        if self.filter.dateranges:
            o = o.filter(Occasion.active_days.op('&&')(array(
                tuple(
                    dt.toordinal()
                    for start, end in self.filter.dateranges
                    for dt in sedate.dtrange(start, end)
                )))
            )

        if self.filter.weekdays:
            o = o.filter(
                Occasion.weekdays.op('&&')(array(self.filter.weekdays)))

        if self.filter.available:
            conditions = []

            for amount in self.filter.available:
                if amount == 'none':
                    conditions.append(Occasion.available_spots == 0)

                elif amount == 'few':
                    conditions.append(Occasion.available_spots.in_((1, 2, 3)))

                elif amount == 'many':
                    conditions.append(Occasion.available_spots >= 4)

            o = o.filter(or_(*conditions))

        # if no filter was applied to the occasion subquery, we ignore it since
        # we would otherwise get zero results
        if o._criterion is not None:
            query = query.filter(model_class.id.in_(o.subquery()))

        return query

    def for_filter(self, **keywords):
        """ Returns a new collection instance.

        The given tag is excluded if already in the list, included if not
        yet in the list. Same goes for the given state.

        Note that dateranges are excluded only if they match exactly. That is
        we don't care about overlaps at all. If the same exact daterange is
        found in the filter, it is excluded.

        """

        return self.__class__(
            session=self.session,
            type=self.type,
            page=0,
            filter=self.filter.toggled(**keywords)
        )

    def by_id(self, id):
        return self.query().filter(Activity.id == id).first()

    def by_name(self, name):
        return self.query().filter(Activity.name == name).first()

    def by_user(self, user):
        return self.query().filter(Activity.username == user.username)

    def by_username(self, username):
        return self.query().filter(Activity.username == username)

    @property
    def used_tags(self):
        """ Returns a list of all the tags used on *all* activites of
        the current type.

        """

        base = self.query_base().with_entities(
            func.skeys(self.model_class._tags).label('keys'))

        query = select([func.array_agg(column('keys'))], distinct=True)
        query = query.select_from(base.subquery())

        tags = self.session.execute(query).scalar()

        return set(tags) if tags else set()

    @property
    def used_municipalities(self):
        """ Returns a list of all the municipalities on *all* activites of
        the current type

        """
        q = self.query_base().with_entities(distinct(Activity.municipality))
        q = q.filter(Activity.municipality != None)

        return set(r[0] for r in q)

    def get_unique_name(self, name):
        """ Given a desired name, finds a variant of that name that's not
        yet used. So if 'foobar' is already used, 'foobar-1' will be returned.

        """
        name = normalize_for_url(name)

        def name_exists(name):
            return self.session.query(exists().where(
                Activity.name == name
            )).scalar()

        for i in range(25):
            if not name_exists(name):
                return name

            name = increment_name(name)

        return secrets.token_hex(8)

    def add(self, title, username, lead=None, text=None, tags=None, name=None):

        type = self.type != '*' and self.type or None

        name = name or self.get_unique_name(title)

        activity = self.model_class(
            name=name,
            title=title,
            tags=tags,
            type=type,
            username=username,
            lead=lead,
            text=text
        )

        self.session.add(activity)
        self.session.flush()

        return activity

    def delete(self, activity):
        for occasion in activity.occasions:
            self.session.delete(occasion)

        self.session.delete(activity)
        self.session.flush()

    def available_weeks(self, period):
        if not period:
            return

        weeknumbers = {n[0] for n in self.session.execute(text("""
            SELECT DISTINCT
                EXTRACT(week FROM start::date)
            FROM OCCASION_DATES
                LEFT JOIN occasions
                ON occasion_id = occasions.id
            WHERE period_id = :period_id
        """), {'period_id': period.id})}

        weeks = sedate.weekrange(period.execution_start, period.execution_end)

        for start, end in weeks:
            if sedate.weeknumber(start) in weeknumbers:
                yield start, end

    def available_ages(self):

        # look at all periods because in the filter view where this is used,
        # we cannot differentiate between the periods unless we make
        # filters interdependent
        query = text("""
            SELECT
                COALESCE(MIN(LOWER(age)), 0),
                COALESCE(MAX(UPPER(age)), 100)
            FROM
                occasions
        """)

        ages = self.session.execute(query).first()

        return ages and ages or None



from datetime import datetime, timedelta
from onegov.activity import ActivityCollection
from onegov.activity import AttendeeCollection
from onegov.activity import BookingCollection
from onegov.activity import OccasionCollection
from onegov.activity import PeriodCollection
from onegov.activity.tests.fixtures.scenario import Scenario
from onegov.core.utils import Bunch, module_path
from onegov.user import User
from onegov.user import UserCollection
from pytest import fixture


@fixture(scope='function')
def owner(session):
    return UserCollection(session).add(
        username='owner@example.org',
        password='hunter2',
        role='editor'
    )


@fixture(scope='function')
def secondary_owner(session):
    return UserCollection(session).add(
        username='secondary@example.org',
        password='hunter2',
        role='editor'
    )


@fixture(scope='function')
def member(session):
    return UserCollection(session).add(
        username='member@example.org',
        password='hunter2',
        role='member'
    )


@fixture(scope='function')
def collections(session):
    return Bunch(
        activities=ActivityCollection(session),
        attendees=AttendeeCollection(session),
        bookings=BookingCollection(session),
        occasions=OccasionCollection(session),
        periods=PeriodCollection(session),
    )


@fixture(scope='function')
def prebooking_period(collections):
    """ Returns a period which is currently in the prebooking phase. """

    s, e = (
        datetime.utcnow() - timedelta(days=10),
        datetime.utcnow() + timedelta(days=10)
    )

    return collections.periods.add(
        title="Testperiod",
        prebooking=(s, e),
        execution=(e + timedelta(days=10), e + timedelta(days=20)),
        active=True
    )


@fixture(scope='function')
def execution_period(collections):
    """ Returns a period which is currently in the execution phase. """

    s, e = (
        datetime.utcnow() - timedelta(days=10),
        datetime.utcnow() + timedelta(days=10)
    )

    return collections.periods.add(
        title="Testperiod",
        prebooking=(s - timedelta(days=20), s - timedelta(days=10)),
        execution=(s, e),
        active=True
    )


@fixture(scope='function')
def inactive_period(collections):
    """ Returns a previously used period """

    s, e = (
        datetime.utcnow() - timedelta(days=100),
        datetime.utcnow() - timedelta(days=10)
    )

    return collections.periods.add(
        title="Testperiod",
        prebooking=(s - timedelta(days=20), s - timedelta(days=10)),
        execution=(s, e),
        active=False
    )


@fixture(scope='session')
def postfinance_xml():
    xml = 'camt.053_P_CH0309000000250090342_380000000_0_2016053100163801.xml'
    xml_path = module_path('onegov.activity', '/tests/fixtures/' + xml)

    with open(xml_path, 'r') as f:
        yield f.read()


@fixture(scope='function')
def scenario(request, session, test_password):

    session.add(User(
        username='admin@example.org',
        password_hash=test_password,
        role='admin'
    ))
    session.add(User(
        username='editor@example.org',
        password_hash=test_password,
        role='editor'
    ))

    yield Scenario(session, test_password)
